# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from collections import defaultdict
from typing import List, Optional, Tuple
import ray
from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy, NodeAffinitySchedulingStrategy
# pylint: disable=unused-import
from ray.util.placement_group import PlacementGroup

from vllm.executor.gpu_executor import GPUExecutor
from vllm.executor.ray_gpu_executor import RayGPUExecutor, RayWorkerWrapper, get_distributed_init_method,\
                                           get_ip, get_vllm_instance_id, get_open_port

from vllm import envs
from vllm.sequence import Logprob, SequenceOutput, SequenceGroupOutput, SamplerOutput, ExecuteModelRequest
from vllm.config import _GB

from llumnix.config import MigrationConfig
from llumnix.logger import init_logger
from llumnix.backends.vllm.utils import get_cache_block_size
from llumnix.backends.profiling import LatencyMemData, SimCacheConfig, model_prefill, model_decode, _pad_to_alignment

logger = init_logger(__name__)

class LlumnixRayGPUExecutor(RayGPUExecutor):
    node_id: str = None
    migration_config: MigrationConfig = None

    def _init_workers_ray(self, placement_group: "PlacementGroup",
                          **ray_remote_kwargs):
        self.last_inference_latency = 0
        if self.parallel_config.tensor_parallel_size == 1:
            # For single GPU case, we use a ray worker with constrained memory.
            num_gpus = self.cache_config.gpu_memory_utilization
        else:
            # Otherwise, the ray workers are allocated with a full GPU.
            num_gpus = 1

        # The driver dummy worker does not actually use any resources.
        # It holds the resource for the driver worker.
        self.driver_dummy_worker: Optional[RayWorkerWrapper] = None
        # The remaining workers are the actual ray actors.
        self.workers: List[RayWorkerWrapper] = []

        if self.parallel_config.ray_workers_use_nsight:
            ray_remote_kwargs = self._configure_ray_workers_use_nsight(
                ray_remote_kwargs)

        # Create the workers.
        driver_ip = get_ip()
        node_id = self.node_id
        for rank in range(self.parallel_config.world_size):
            if placement_group:
                bundle = placement_group.bundle_specs[rank+1]
                if not bundle.get("GPU", 0):
                    raise Exception("GPU resource cannot be 0.")
                scheduling_strategy = PlacementGroupSchedulingStrategy(
                    placement_group=placement_group,
                    placement_group_capture_child_tasks=True,
                )
            else:
                scheduling_strategy = NodeAffinitySchedulingStrategy(
                    node_id=node_id,
                    soft=False,
                )
            worker = ray.remote(
                num_cpus=0,
                num_gpus=num_gpus,
                scheduling_strategy=scheduling_strategy,
                max_concurrency=2,
                **ray_remote_kwargs,
            )(RayWorkerWrapper).remote(
                worker_module_name="llumnix.backends.vllm.worker",
                worker_class_name="MigrationWorker",
                trust_remote_code=self.model_config.trust_remote_code,
            )
            worker_ip = ray.get(worker.get_node_ip.remote())
            if worker_ip == driver_ip and self.driver_dummy_worker is None:
                # If the worker is on the same node as the driver, we use it
                # as the resource holder for the driver process.
                self.driver_dummy_worker = worker
                self.driver_worker = RayWorkerWrapper(
                    worker_module_name="llumnix.backends.vllm.worker",
                    worker_class_name="MigrationWorker",
                    trust_remote_code=self.model_config.trust_remote_code,
                )
            else:
                # Else, added to the list of workers.
                self.workers.append(worker)

        if self.driver_dummy_worker is None:
            raise ValueError(
                "Ray does not allocate any GPUs on the driver node. Consider "
                "adjusting the Ray placement group or running the driver on a "
                "GPU node.")

        # Get the set of GPU IDs used on each node.
        worker_node_and_gpu_ids = self._run_workers("get_node_and_gpu_ids",
                                                    use_dummy_driver=True)

        node_workers = defaultdict(list)
        node_gpus = defaultdict(list)

        for i, (node_id, gpu_ids) in enumerate(worker_node_and_gpu_ids):
            node_workers[node_id].append(i)
            node_gpus[node_id].extend(gpu_ids)
        for node_id, gpu_ids in node_gpus.items():
            node_gpus[node_id] = sorted(gpu_ids)

        # pylint: disable=invalid-name
        VLLM_INSTANCE_ID = get_vllm_instance_id()

        # Set environment variables for the driver and workers.
        all_args_to_update_environment_variables = [({
            "CUDA_VISIBLE_DEVICES":
            ",".join(map(str, node_gpus[node_id])),
            "VLLM_INSTANCE_ID":
            VLLM_INSTANCE_ID,
            "VLLM_TRACE_FUNCTION":
            str(envs.VLLM_TRACE_FUNCTION),
        }, ) for (node_id, _) in worker_node_and_gpu_ids]
        self._run_workers("update_environment_variables",
                          all_args=all_args_to_update_environment_variables)

        distributed_init_method = get_distributed_init_method(
            driver_ip, get_open_port())

        # Initialize the actual workers inside worker wrapper.
        init_worker_all_kwargs = [
            self._get_worker_kwargs(
                local_rank=node_workers[node_id].index(rank),
                rank=rank,
                distributed_init_method=distributed_init_method,
            ) for rank, (node_id, _) in enumerate(worker_node_and_gpu_ids)
        ]
        self._run_workers("init_worker", all_kwargs=init_worker_all_kwargs)

        self._run_workers("init_device")
        self._run_workers("load_model",
                          max_concurrent_workers=self.parallel_config.
                          max_parallel_loading_workers)
        self._run_workers("reserve_memory_for_migration",
                          migration_config=self.migration_config,
                          model_config=self.model_config,
                          cache_config=self.cache_config,
                          parallel_config=self.parallel_config)

    def execute_model(self, *args, **kwargs):
        t0 = time.time()
        outputs = super().execute_model(*args, **kwargs)
        t1 = time.time()
        self.last_inference_latency = (t1 - t0) * 1000
        return outputs

class SimGPUExecutor(GPUExecutor):
    latency_mem: LatencyMemData = None
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.last_inference_latency = 0
        self.migration_bandwidth = self.latency_mem.migration_bandwidth
        # TODO(ZeldaHuang): add swap bandwidth

        self.cache_block_size = get_cache_block_size(
            self.cache_config.block_size, self.model_config, self.parallel_config)
        self.cache_block_size /= _GB
        self.sim_cache_config = SimCacheConfig(self.cache_config.gpu_memory_utilization,
                                               self.cache_config.block_size,
                                               self.scheduler_config.max_num_batched_tokens)

    def _init_executor(self) -> None:
        pass

    def determine_num_available_blocks(self) -> Tuple[int, int]:
        num_gpu_blocks = self.latency_mem.cache_dict.get(self.sim_cache_config, 880)
        num_cpu_blocks = 2048
        return (num_gpu_blocks, num_cpu_blocks)

    def initialize_cache(self, num_gpu_blocks: int,
                         num_cpu_blocks: int) -> None:
        logger.info("# GPU blocks: %d, # CPU blocks: %d", num_gpu_blocks, num_cpu_blocks)

    def execute_model(
            self,
            execute_model_req: ExecuteModelRequest) -> List[SamplerOutput]:
        prefill_seq_len = 0
        decode_seq_len = 0
        decode_bs = 0
        for meta_data in execute_model_req.seq_group_metadata_list:
            if meta_data.is_prompt:
                prefill_seq_len += meta_data.token_chunk_size
            else:
                decode_bs += meta_data.token_chunk_size
                decode_seq_len += list(meta_data.seq_data.values())[0].get_len()
        decode_bs = _pad_to_alignment(decode_bs, 8)
        latency = 0
        if prefill_seq_len:
            latency += self.latency_mem.prefill_latency[prefill_seq_len][0] if prefill_seq_len in self.latency_mem.prefill_latency \
                       else model_prefill(prefill_seq_len, *self.latency_mem.prefill_model_params)
        if decode_bs:
            decode_meta_data = (decode_bs, decode_seq_len)
            latency += self.latency_mem.decode_latency[decode_meta_data][0] if decode_meta_data in self.latency_mem.decode_latency \
                       else model_decode((decode_bs, decode_seq_len), *self.latency_mem.decode_model_params)
        time.sleep(latency/1000)
        sampler_outputs = []
        for meta_data in execute_model_req.seq_group_metadata_list:
            samples = []
            for seq_id in meta_data.seq_data.keys():
                dummy_sample_output = SequenceOutput(seq_id, 20, {20: Logprob(1.0)})
                samples.append(dummy_sample_output)
            if samples:
                output = SequenceGroupOutput(samples, None)
                sampler_outputs.append(output)
        return [SamplerOutput(outputs=sampler_outputs)]

    def send_blocks(self, blocks_len) -> None:
        migration_latency = (self.cache_block_size * blocks_len) / self.migration_bandwidth
        time.sleep(migration_latency)



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from vllm.sequence import Sequence
from vllm.sequence import Logprob
from llumnix.backends.vllm.scheduler import BlockManagerLlumnix
from .utils import create_dummy_prompt, initialize_scheduler


def get_sequence_groups(scheduler_output):
    return [s.seq_group for s in scheduler_output.scheduled_seq_groups]

def schedule_and_update_computed_tokens(scheduler):
    metas, out = scheduler.schedule()
    for s, meta in zip(out.scheduled_seq_groups, metas):
        s.seq_group.update_num_computed_tokens(meta.token_chunk_size)
    return metas, out

def append_new_token(out, token_id: int):
    seq_groups = get_sequence_groups(out)
    for seq_group in seq_groups:
        for seq in seq_group.get_seqs():
            seq.append_token_id(token_id, {token_id: Logprob(token_id)})

def test_manager_get_free_blocks():
    block_size = 4
    num_cpu_blocks = 4
    num_gpu_blocks = 4
    block_manager = BlockManagerLlumnix(block_size,
                                        num_cpu_blocks,
                                        num_gpu_blocks,
                                        watermark=0)
    before_allocate = block_manager.get_num_free_gpu_blocks()
    block_table = block_manager.get_free_blocks(2)
    after_allocate = block_manager.get_num_free_gpu_blocks()
    assert after_allocate + 2 == before_allocate
    block_manager._free_block_table(block_table)
    after_free = block_manager.get_num_free_gpu_blocks()
    assert after_free == before_allocate

def test_manager_add_block_table():
    block_size = 4
    num_cpu_blocks = 4
    num_gpu_blocks = 4
    block_manager = BlockManagerLlumnix(block_size,
                                        num_cpu_blocks,
                                        num_gpu_blocks,
                                        watermark=0)
    block_table = block_manager.get_free_blocks(2)
    seq = Sequence(1,"1",[0], block_size=block_size)
    block_manager.add_block_table(block_table, seq.seq_id)
    after_allocate = block_manager.get_num_free_gpu_blocks()
    assert after_allocate + 2 == num_gpu_blocks
    block_manager.free(seq)
    after_free = block_manager.get_num_free_gpu_blocks()
    assert after_free == num_gpu_blocks

def test_scheduler_policy():
    scheduler = initialize_scheduler()
    num_seq_group = 4
    block_size = 4
    for idx in range(1, num_seq_group + 1):
        _, seq_group = create_dummy_prompt(str(idx), prompt_length=idx, block_size=block_size)
        scheduler.add_seq_group(seq_group)

    # all seq_group in waiting queue
    migrating_request = scheduler.get_last_running_request()
    assert migrating_request is None
    migrating_request = scheduler.get_shortest_running_request()
    assert migrating_request is None
    migrating_request = scheduler.get_longest_running_request()
    assert migrating_request is None
    # all seq_group in prefilling stage
    _, out = schedule_and_update_computed_tokens(scheduler)
    migrating_request = scheduler.get_last_running_request()
    assert migrating_request is None
    migrating_request = scheduler.get_shortest_running_request()
    assert migrating_request is None
    migrating_request = scheduler.get_longest_running_request()
    assert migrating_request is None
    append_new_token(out, 1)
    schedule_and_update_computed_tokens(scheduler)
    # all in running queue
    migrating_request = scheduler.get_last_running_request()
    assert migrating_request.request_id == str(num_seq_group)
    migrating_request = scheduler.get_shortest_running_request()
    assert migrating_request.request_id == "1"
    migrating_request = scheduler.get_longest_running_request()
    assert migrating_request.request_id == str(num_seq_group)

def test_scheduler_num_killed_request():
    scheduler = initialize_scheduler()
    # tot 8 blocks
    num_seq_group = 4
    block_size = 4
    for idx in range(1, num_seq_group + 1):
        _, seq_group = create_dummy_prompt(str(idx), prompt_length=8, block_size=block_size)
        scheduler.add_seq_group(seq_group)
    # remain 0 blocks
    _, out = schedule_and_update_computed_tokens(scheduler)
    append_new_token(out, 1)
    assert scheduler._get_num_killed_requests() == 0
    # preempt 2 requests
    _, out = schedule_and_update_computed_tokens(scheduler)
    assert scheduler._get_num_killed_requests() == 2

def test_scheduler_running_request():
    scheduler = initialize_scheduler()
    num_seq_group = 4
    block_size = 4
    for idx in range(1, num_seq_group + 1):
        _, seq_group = create_dummy_prompt(str(idx), prompt_length=idx, block_size=block_size)
        scheduler.add_seq_group(seq_group)
    schedule_and_update_computed_tokens(scheduler)
    assert scheduler.get_num_unfinished_seq_groups() == 4
    scheduler.remove_running_request("1")
    assert scheduler.get_num_unfinished_seq_groups() == 3
    _, seq_group = create_dummy_prompt("5", prompt_length=idx, block_size=block_size)
    scheduler.add_running_request(seq_group)
    assert scheduler.get_num_unfinished_seq_groups() == 4

def test_scheduler_migrating_out_request_last_stage():
    scheduler = initialize_scheduler()
    block_size = 4
    _, seq_group = create_dummy_prompt("1", prompt_length=1, block_size=block_size)
    scheduler.add_migrating_out_request_last_stage(seq_group)
    assert len(scheduler.pop_migrating_out_requests_last_stage()) == 1
    scheduler.add_migrating_out_request_last_stage(seq_group)
    scheduler.remove_migrating_out_request_last_stage(seq_group)
    assert len(scheduler.pop_migrating_out_requests_last_stage()) == 0

def test_scheduler_pre_alloc():
    # total 8 blocks
    scheduler = initialize_scheduler()
    blocks = scheduler.pre_alloc("1", 2)
    assert len(blocks) == 2
    assert len(scheduler.pre_alloc_cache_dict["1"]) == 2
    blocks = scheduler.pre_alloc("1", 4)
    assert len(blocks) == 4
    assert len(scheduler.pre_alloc_cache_dict["1"]) == 6
    blocks = scheduler.pre_alloc("2,", 4)
    assert len(blocks) == 0

def test_scheduler_should_abort_migration():
    scheduler = initialize_scheduler()
    # tot 8 blocks
    block_size = 4
    _, seq_group_0 = create_dummy_prompt("0", prompt_length=7, block_size=block_size)
    scheduler.add_seq_group(seq_group_0)
    _, seq_group_1 = create_dummy_prompt("1", prompt_length=17, block_size=block_size)
    scheduler.add_seq_group(seq_group_1)
    # remain 0 blocks
    _, out = schedule_and_update_computed_tokens(scheduler)
    append_new_token(out, 1)

    assert scheduler._get_num_killed_requests() == 0
    # assert scheduler.block_manager.get_num_free_gpu_blocks() == 0
    # all in running queue
    _, out = schedule_and_update_computed_tokens(scheduler)
    append_new_token(out, 1)
    assert scheduler._get_num_killed_requests() == 0
    migrating_request = scheduler.get_last_running_request()
    last_stage_time = time.time()
    assert migrating_request.request_id == "1"
    # preempt request 1
    _, out = schedule_and_update_computed_tokens(scheduler)
    append_new_token(out, 1)
    assert scheduler.should_abort_migration(seq_group_1, last_stage_time)
    assert not scheduler.should_abort_migration(seq_group_0, last_stage_time)
    assert scheduler._get_num_killed_requests() == 1
    scheduler.remove_running_request(seq_group_0)
    scheduler.free_src_request(seq_group_0)
    # free request 0, requset 1 prefill
    _, out = schedule_and_update_computed_tokens(scheduler)
    append_new_token(out, 1)
    assert scheduler._get_num_killed_requests() == 0
    assert scheduler.should_abort_migration(seq_group_1, last_stage_time)

def test_free_dst_pre_alloc_cache():
    scheduler = initialize_scheduler()
    scheduler.pre_alloc("1", 2)
    scheduler.pre_alloc("1", 4)
    assert len(scheduler.pre_alloc_cache_dict["1"]) == 6
    scheduler.free_dst_pre_alloc_cache("1")
    assert scheduler.pre_alloc_cache_dict.get("1", None) is None
    assert scheduler.block_manager.get_num_free_gpu_blocks() == 8

def test_get_request_incremental_blocks():
    scheduler = initialize_scheduler()
    block_size = 4
    _, seq_group = create_dummy_prompt("0", prompt_length=16, block_size=block_size)
    scheduler.add_seq_group(seq_group)
    schedule_and_update_computed_tokens(scheduler)
    incremental_blocks = scheduler.get_request_incremental_blocks(seq_group, 2)
    assert len(incremental_blocks) == 2



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import List, Optional, Dict, Union, Iterable, Any
from collections import defaultdict
import threading
import ray
# pylint: disable=unused-import
from ray.util.placement_group import PlacementGroup
from ray.util.queue import Queue as RayQueue

from vllm.engine.llm_engine import LLMEngine
from vllm.core.scheduler import ScheduledSequenceGroup
from vllm.outputs import RequestOutput
from vllm.sequence import SequenceGroup, SequenceStatus, SamplerOutput, SequenceGroupMetadata
from vllm.engine.arg_utils import EngineArgs
from vllm.utils import Counter
from vllm.usage.usage_lib import UsageContext

from llumnix.logger import init_logger
from llumnix.llumlet.migrating_request import MigratingRequest
from llumnix.instance_info import InstanceInfo
from llumnix.backends.backend_interface import BackendInterface
from llumnix.backends.vllm.scheduler import SchedulerLlumnix
from llumnix.backends.vllm.utils import detect_unsupported_feature
from llumnix.backends.profiling import LatencyMemData
from llumnix.server_info import ServerInfo
from llumnix.config import MigrationConfig


logger = init_logger(__name__)

class LLMEngineLlumnix(LLMEngine):
    def __init__(self, instance_id: str, *arg, **kwargs) -> None:
        super().__init__(*arg, **kwargs)
        self.instance_id = instance_id
        self.step_counter = Counter()
        self.instance_info = None
        self.scaling_down = False
        self.request_server_info: Dict[str, ServerInfo] = {}

    # pylint: disable=W0221
    @classmethod
    def from_engine_args(
        cls,
        engine_args: EngineArgs,
        migration_config: MigrationConfig,
        usage_context: UsageContext = UsageContext.ENGINE_CONTEXT,
        instance_id: str = None,
        placement_group: Optional["PlacementGroup"] = None,
        node_id: str = None,
        latency_mem: Optional[LatencyMemData] = None
    ) -> "LLMEngineLlumnix":
        """Creates an LLM engine from the engine arguments."""
        # Create the engine configs.
        detect_unsupported_feature(engine_args)
        engine_config = engine_args.create_engine_config()
        engine_config.parallel_config.placement_group = placement_group
        # Initialize the cluster and specify the executor class.
        # pylint: disable=import-outside-toplevel
        if latency_mem is not None:
            from llumnix.backends.vllm.executor import SimGPUExecutor
            executor_class = SimGPUExecutor
            executor_class.latency_mem = latency_mem
        elif engine_config.parallel_config.worker_use_ray:
            from llumnix.backends.vllm.executor import LlumnixRayGPUExecutor
            executor_class = LlumnixRayGPUExecutor
            executor_class.migration_config = migration_config
        else:
            raise ValueError('Unsupported executor backend')
        # Hack to pass node_id to _init_workers_ray function.
        executor_class.node_id = node_id
        # Create the LLM engine.
        engine = cls(
            instance_id=instance_id,
            **engine_config.to_dict(),
            executor_class=executor_class,
            log_stats=not engine_args.disable_log_stats,
            usage_context=usage_context,
        )
        return engine

    def _process_model_outputs(
        self,
        output: List[SamplerOutput],
        scheduled_seq_groups: List[ScheduledSequenceGroup],
        ignored_seq_groups: List[SequenceGroup],
        seq_group_metadata_list: List[SequenceGroupMetadata],
    ) -> List[RequestOutput]:
        # ensure scheduled_seq_groups matching output
        with self.scheduler.scheduler_lock:
            if output:
                new_output = []
                new_scheduled_seq_groups = []
                new_seq_group_metadata_list = []
                for scheduled_seq_group, seq_group_meta, seq_group_output in zip(scheduled_seq_groups, seq_group_metadata_list, output[0].outputs):
                    seq_group = scheduled_seq_group.seq_group
                    if seq_group.get_seqs(SequenceStatus.RUNNING):
                        new_scheduled_seq_groups.append(scheduled_seq_group)
                        new_seq_group_metadata_list.append(seq_group_meta)
                        new_output.append(seq_group_output)
                scheduled_seq_groups = new_scheduled_seq_groups
                output[0].outputs = new_output
                seq_group_metadata_list = new_seq_group_metadata_list
            return super()._process_model_outputs(output, scheduled_seq_groups, ignored_seq_groups, seq_group_metadata_list)

    def step(self) -> None:
        output_list = super().step()

        instance_info: InstanceInfo = self.instance_info

        if self.scaling_down:
            instance_info.num_running_requests = 1
            instance_info.num_available_gpu_blocks = -self.cache_config.num_gpu_blocks
            instance_info.num_available_gpu_blocks_waiting = -self.cache_config.num_gpu_blocks

        instance_info.instance_id = self.instance_id
        instance_info.step_id = next(self.step_counter)
        instance_info.timestamp = time.time()
        instance_info.latency = self.model_executor.last_inference_latency
        seq_groups = self.scheduler.running
        if seq_groups:
            tot_blocks = []
            for seq in seq_groups[-1].get_seqs(SequenceStatus.RUNNING):
                blocks = self.scheduler.block_manager.get_block_table(seq)
                tot_blocks.extend(blocks)
            tot_blocks = set(tot_blocks)
            instance_info.num_blocks_last_running_request = len(tot_blocks)

        self.free_request_states(instance_info.finished_request_ids)

        if len(output_list) > 0:
            server_info_list = []
            for output in output_list:
                server_info_list.append(self.request_server_info[output.request_id])
            self._put_request_output_to_server(output_list, server_info_list)
        self.instance_info = instance_info

    def update_instance_info(self, instance_info: InstanceInfo) -> None:
        # These fields are updated after step.
        if self.instance_info is not None:
            instance_info.instance_id = self.instance_info.instance_id
            instance_info.step_id = self.instance_info.step_id
            instance_info.timestamp = self.instance_info.timestamp
            instance_info.latency = self.instance_info.latency
            instance_info.num_blocks_last_running_request = self.instance_info.num_blocks_last_running_request
        self.instance_info = instance_info

    def _put_request_output_to_server(self,
                                      request_outputs: List[RequestOutput],
                                      server_infos: List[ServerInfo]) -> None:
        server_request_outputs = defaultdict(list)
        server_queue: Dict[str, RayQueue] = {}
        # Reorganize data in orther to put request output to queue in batch at one time.
        for request_output, server_info in zip(request_outputs, server_infos):
            server_id = server_info.server_id
            request_output_queue = server_info.request_output_queue
            server_request_outputs[server_id].append(request_output)
            if server_id not in server_queue:
                server_queue[server_id] = request_output_queue
        for server_id, req_outputs in server_request_outputs.items():
            try:
                server_queue[server_id].actor.put_nowait_batch.remote(req_outputs)
            except ray.exceptions.RayActorError:
                logger.info("Server {} is dead".format(server_id))
                request_ids = [req_output.request_id for req_output in req_outputs]
                self.abort(request_ids)

    def free_request_states(self, request_id: Union[str, Iterable[str]]) -> None:
        if isinstance(request_id, str):
            request_id = (request_id,)
        request_ids = set(request_id)
        for req_id in request_ids:
            if req_id in self.request_server_info:
                del self.request_server_info[req_id]
            if req_id in self.scheduler.last_preemption_time_dict:
                del self.scheduler.last_preemption_time_dict[req_id]

class BackendVLLM(BackendInterface):
    def __init__(
        self,
        instance_id: str,
        migration_config: MigrationConfig,
        engine_args: EngineArgs,
        placement_group: "PlacementGroup" = None,
        node_id: str = None
    ) -> None:
        self.engine: LLMEngineLlumnix = LLMEngineLlumnix.from_engine_args(engine_args=engine_args,
                                                                          migration_config=migration_config,
                                                                          instance_id=instance_id,
                                                                          placement_group=placement_group,
                                                                          node_id=node_id)
        # multi-instance args
        self.engine.scheduler = SchedulerLlumnix(self.engine.scheduler_config, self.engine.cache_config, self.engine.lora_config)
        self.engine.scheduler.add_update_instance_info_callback(self.engine.update_instance_info)
        self.engine.output_processor.scheduler = self.engine.scheduler
        self.instance_id = instance_id
        self.worker_handle_list = self.engine.model_executor.workers.copy()
        if len(self.worker_handle_list) + 1 == self.engine.parallel_config.world_size:
            self.worker_handle_list.insert(0, ray.get_actor(f"instance_{self.instance_id}", namespace="llumnix"))
        self._run_workers("init_migration", instance_id=instance_id, migration_config=migration_config,\
                                                      src_worker_handle_list=self.worker_handle_list,
                                                      placement_group=placement_group, node_id=node_id)
        self._thread = threading.Thread(
            target=self._start_engine_loop, args=(), daemon=True, name="engine_loop"
        )
        self._thread.start()

    def _start_engine_loop(self) -> None:
        while True:
            self.engine.step()

    def execute_worker_method(self, method, *args, **kwargs):
        return self.engine.model_executor.driver_worker.execute_method(method, *args, **kwargs)

    def stop_shutdown(self) -> None:
        self.engine.scaling_down = False

    def shutdown_workers(self):
        migrated_requests = []

        self.engine.scaling_down = True
        while self.has_unfinished_requests() and self.engine.scaling_down:
            time.sleep(1)
        time.sleep(0.1)
        if self.engine.scaling_down:
            self._run_workers(
                "shutdown",
                )
        return migrated_requests

    def restart_workers(self) -> None:
        self._run_workers(
            "restart",
            )
        self.engine.scaling_down = False

    def add_request(self,
                    request_id: str,
                    server_info: ServerInfo,
                    *args,
                    **kwargs) -> None:
        # Store the server information of each request to put the request outputs back to the corresponding api server correctly.
        self.engine.request_server_info[request_id] = server_info
        self.engine.add_request(request_id, *args, **kwargs)

    def commit_dst_request(self, backend_request: SequenceGroup, server_info: ServerInfo) -> None:
        seq = backend_request.get_seqs()[0]
        seq.seq_id = next(self.engine.seq_counter)
        logger.info("add seq {} to block table".format(seq.seq_id))
        pre_alloc_blocks = self.engine.scheduler.pre_alloc_cache_dict.pop(backend_request.request_id)
        self.engine.scheduler.block_manager.add_block_table(pre_alloc_blocks, seq.seq_id)
        self.add_running_request(backend_request)
        self.engine.request_server_info[backend_request.request_id] = server_info

    def send_blocks(self, dst_ray_actor: "ray.actor.ActorHandle", src_blocks: List[int], dst_blocks: List[int]) -> None:
        ray.get(dst_ray_actor.execute_engine_method.remote("_run_workers",
                                                           "migrate_cache",
                                                           dst_blocks=dst_blocks,
                                                           src_blocks=src_blocks,
                                                           src_worker_handle_list=self.worker_handle_list))

    def _run_workers(self, *args, **kwargs):
        # pylint: disable=protected-access
        return self.engine.model_executor._run_workers(*args, **kwargs)

    def is_ready(self):
        return True

    def abort_request(self, request_id: Union[str, Iterable[str]]) -> None:
        if isinstance(request_id, str):
            request_id = (request_id,)
        request_ids = set(request_id)
        self.free_request_states(request_ids)
        return self.engine.abort_request(request_ids)

    def free_request_states(self, request_id: Union[str, Iterable[str]]) -> None:
        return self.engine.free_request_states(request_id)

    def get_request_incremental_blocks(self, *args, **kwargs) -> List[int]:
        return self.engine.scheduler.get_request_incremental_blocks(*args, **kwargs)

    def remove_running_request(self, *args, **kwargs) -> None:
        return self.engine.scheduler.remove_running_request(*args, **kwargs)

    def add_migrating_out_request_last_stage(self, *args, **kwargs) -> None:
        return self.engine.scheduler.add_migrating_out_request_last_stage(*args, **kwargs)

    def remove_migrating_out_request_last_stage(self, *args, **kwargs) -> None:
        return self.engine.scheduler.remove_migrating_out_request_last_stage(*args, **kwargs)

    def pop_migrating_out_requests_last_stage(self, *args, **kwargs) -> List[Any]:
        return self.engine.scheduler.pop_migrating_out_requests_last_stage(*args, **kwargs)

    def pre_alloc(self, *args, **kwargs) -> List[int]:
        return self.engine.scheduler.pre_alloc(*args, **kwargs)

    def should_abort_migration(self, *args, **kwargs) -> bool:
        return self.engine.scheduler.should_abort_migration(*args, **kwargs)

    def add_running_request(self, *args, **kwargs) -> None:
        return self.engine.scheduler.add_running_request(*args, **kwargs)

    def is_request_running(self, *args, **kwargs) -> bool:
        return self.engine.scheduler.is_request_running(*args, **kwargs)

    def free_dst_pre_alloc_cache(self, *args, **kwargs) -> None:
        return self.engine.scheduler.free_dst_pre_alloc_cache(*args, **kwargs)

    def free_src_request(self, backend_request: SequenceGroup) -> None:
        return self.engine.scheduler.free_src_request(backend_request)

    def get_last_running_request(self) -> Optional[MigratingRequest]:
        return self.engine.scheduler.get_last_running_request()

    def get_longest_running_request(self) -> Optional[MigratingRequest]:
        return self.engine.scheduler.get_longest_running_request()

    def get_shortest_running_request(self) -> Optional[MigratingRequest]:
        return self.engine.scheduler.get_shortest_running_request()

    def get_request_server_info(self, request_id: str) -> ServerInfo:
        return self.engine.request_server_info[request_id]

    def get_all_request_ids(self) -> List[str]:
        return self.engine.scheduler.get_all_request_ids()



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import pytest

from llumnix.instance_info import InstanceLoadCalculator, InstanceInfo
from llumnix.global_scheduler.dispatch_scheduler import DispatchScheduler


def init_dispatch_scheduler(policy='load'):
    instance_load_calculator = InstanceLoadCalculator('remaining_steps', True)
    dispatch_scheduler = DispatchScheduler(policy, instance_load_calculator)
    return dispatch_scheduler

@pytest.fixture
def dispatch_scheduler():
    dispatch_scheduler = init_dispatch_scheduler()
    yield dispatch_scheduler

def test_add_instance_and_remove_instance(dispatch_scheduler):
    dispatch_scheduler.add_instance('instance_1')
    assert dispatch_scheduler.num_instances == 1
    dispatch_scheduler.add_instance('instance_2')
    assert dispatch_scheduler.num_instances == 2
    dispatch_scheduler.remove_instance('instance_1')
    assert dispatch_scheduler.num_instances == 1
    dispatch_scheduler.remove_instance('instance_2')
    assert dispatch_scheduler.num_instances == 0

def test_dispatch_balanced():
    dispatch_scheduler = init_dispatch_scheduler('balanced')
    num_tests = 100
    for _ in range(num_tests):
        instance_num_requests = {}
        for instance_id in ['instance_1', 'instance_2', 'instance_3', 'instance_4']:
            instance_num_requests[instance_id] = random.randint(1, 10)
        dispatch_scheduler.instance_num_requests = instance_num_requests
        min_instance_id = next(key for key, value in sorted(instance_num_requests.items(), key=lambda item: item[1]))
        instance_id = dispatch_scheduler.dispatch()
        assert min_instance_id == instance_id

def test_dispatch_load():
    dispatch_scheduler = init_dispatch_scheduler('load')
    num_tests = 100
    for _ in range(num_tests):
        instance_num_requests = {}
        instance_info_dict = {}
        for instance_id in ['instance_1', 'instance_2', 'instance_3', 'instance_4']:
            instance_num_requests[instance_id] = 0
            instance_info = InstanceInfo()
            instance_info.instance_id = instance_id
            instance_info.instance_load_dispatch_scale = random.random()
            instance_info_dict[instance_id] = instance_info
        dispatch_scheduler.instance_num_requests = instance_num_requests
        dispatch_scheduler.instance_info = instance_info_dict
        min_instance_id = next(key for key, value in sorted(instance_info_dict.items(),
                                                            key=lambda item: item[1].instance_load_dispatch_scale))
        instance_id = dispatch_scheduler.dispatch()
        assert min_instance_id == instance_id

def test_dispatch_queue():
    dispatch_scheduler = init_dispatch_scheduler('queue')
    num_tests = 100
    for _ in range(num_tests):
        instance_num_requests = {}
        instance_info_dict = {}
        for instance_id in ['instance_1', 'instance_2', 'instance_3', 'instance_4']:
            instance_num_requests[instance_id] = 0
            instance_info = InstanceInfo()
            instance_info.instance_id = instance_id
            instance_info.num_waiting_requests = random.randint(1, 10)
            instance_info_dict[instance_id] = instance_info
        dispatch_scheduler.instance_num_requests = instance_num_requests
        dispatch_scheduler.instance_info = instance_info_dict
        min_instance_id = next(key for key, value in sorted(instance_info_dict.items(),
                                                            key=lambda item: item[1].num_waiting_requests))
        instance_id = dispatch_scheduler.dispatch()
        assert instance_info_dict[min_instance_id].num_waiting_requests == instance_info_dict[instance_id].num_waiting_requests



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock

from vllm.sequence import (Logprob, SequenceGroupOutput, SequenceOutput,
                           SequenceStatus,SamplerOutput)
from vllm import EngineArgs
from vllm.engine.output_processor.single_step import SingleStepOutputProcessor
from vllm.engine.output_processor.stop_checker import StopChecker
from vllm.transformers_utils.detokenizer import Detokenizer
from vllm.utils import Counter

from llumnix.backends.vllm.llm_engine import LLMEngineLlumnix
from llumnix.backends.vllm.executor import LlumnixRayGPUExecutor, SimGPUExecutor
from llumnix.backends.profiling import LatencyMemData

from .utils import create_dummy_prompt, initialize_scheduler


class MockEngine(LLMEngineLlumnix):
    def __init__(self, executor_class=None, *args, **kwargs):
        self.scheduler = initialize_scheduler()
        detokenizer = MagicMock(spec=Detokenizer)
        stop_checker = MagicMock(spec=StopChecker)
        seq_counter = Counter()
        self.executor_class = executor_class

        self.output_processor = SingleStepOutputProcessor(self.scheduler.scheduler_config,detokenizer, self.scheduler, seq_counter, stop_checker)


def test_llm_engine_process_model_outputs():

    llm_engine = MockEngine()
    _, seq_group_0 = create_dummy_prompt(
        "0", prompt_length=7, block_size=4
    )
    _, seq_group_1 = create_dummy_prompt(
        "1", prompt_length=7, block_size=4
    )
    llm_engine.scheduler.add_seq_group(seq_group_0)
    llm_engine.scheduler.add_seq_group(seq_group_1)
    metas, out = llm_engine.scheduler.schedule()

    seqs = [seq_group_0.get_seqs()[0], seq_group_1.get_seqs()[0]]

    outputs = [
        SequenceGroupOutput(
            samples=[
                SequenceOutput(
                    parent_seq_id=seq.seq_id,
                    output_token=1,
                    logprobs={1: Logprob(0.0)},
                )
            ],
            prompt_logprobs=None,
        ) for seq in seqs
    ]
    sampler_outputs = [SamplerOutput(outputs=outputs)]

    scheduled_seq_groups = out.scheduled_seq_groups
    # normal case, all requests be processed
    ret = llm_engine._process_model_outputs(sampler_outputs, scheduled_seq_groups,[], metas)
    assert len(ret) == 2
    metas, out = llm_engine.scheduler.schedule()
    scheduled_seq_groups = out.scheduled_seq_groups
    seqs[0].status=SequenceStatus.WAITING
    # migration case , requests stopping during last stage migration, stop process
    ret = llm_engine._process_model_outputs(sampler_outputs, scheduled_seq_groups,[], metas)
    assert len(ret) == 1

def test_llm_engine_from_engine_args():
    engine_args = EngineArgs(model="facebook/opt-125m", worker_use_ray=True)
    llm_engine = MockEngine.from_engine_args(engine_args, instance_id="0", migration_config=None)
    assert llm_engine.executor_class == LlumnixRayGPUExecutor

    latency_data = LatencyMemData({},{},{})
    llm_engine = MockEngine.from_engine_args(engine_args, instance_id="0", migration_config=None, latency_mem=latency_data)
    assert llm_engine.executor_class == SimGPUExecutor



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import pytest
import torch
import ray

from vllm.engine.arg_utils import EngineArgs

from llumnix.backends.vllm.worker import MigrationWorker
from llumnix.arg_utils import EngineManagerArgs
from llumnix.utils import random_uuid

from tests.backends.vllm.test_worker import create_worker

class MockMigrationWorker(MigrationWorker):
    def set_gpu_cache(self, data):
        for layer_idx in range(self.cache_engine.num_layers):
            self.gpu_cache[layer_idx].copy_(data[layer_idx])
        torch.cuda.synchronize()

    def get_gpu_cache(self):
        torch.cuda.synchronize()
        return self.gpu_cache

@pytest.mark.skipif(torch.cuda.device_count() < 2, reason="Need at least 2 GPU to run the test.")
@pytest.mark.parametrize("backend", ['rpc', 'gloo', 'nccl'])
def test_migrate_cache(backend):
    ray.init(namespace="llumnix", ignore_reinit_error=True)

    engine_config = EngineArgs(model='facebook/opt-125m', max_model_len=8, enforce_eager=True).create_engine_config()
    migraiton_config = EngineManagerArgs(migration_cache_blocks=3, migration_num_layers=5).create_migration_config()
    migraiton_config.migration_backend = backend

    worker0 = create_worker(rank=0, local_rank=0, engine_config=engine_config,
                            worker_module_name="tests.backends.vllm.test_migration_backend",
                            worker_class_name="MockMigrationWorker")
    worker1 = create_worker(rank=0, local_rank=0, engine_config=engine_config,
                            worker_module_name="tests.backends.vllm.test_migration_backend",
                            worker_class_name="MockMigrationWorker")

    ray.get(worker0.execute_method.remote('init_device'))
    ray.get(worker1.execute_method.remote('init_device'))

    num_gpu_blocks = 8
    ray.get(worker0.execute_method.remote('initialize_cache', num_gpu_blocks=num_gpu_blocks, num_cpu_blocks=0))
    ray.get(worker1.execute_method.remote('initialize_cache', num_gpu_blocks=num_gpu_blocks, num_cpu_blocks=0))

    worker0_id = random_uuid()
    ray.get(worker0.execute_method.remote(
        'init_migration',
        instance_id=worker0_id,
        migration_config=migraiton_config,
        src_worker_handle_list=[worker0],
        node_id=ray.get_runtime_context().get_node_id()))

    worker1_id = random_uuid()
    ray.get(worker1.execute_method.remote(
        'init_migration',
        instance_id=worker1_id,
        migration_config=migraiton_config,
        src_worker_handle_list=[worker1],
        node_id=ray.get_runtime_context().get_node_id()))

    instance_rank = {worker0_id: 0, worker1_id: 1}
    group_name = random_uuid()
    assert all(ray.get([worker0.execute_method.remote('rebuild_migration_backend',
                                        instance_rank=instance_rank, group_name=group_name),
                        worker1.execute_method.remote('rebuild_migration_backend',
                                        instance_rank=instance_rank, group_name=group_name)]))
    assert all(ray.get([worker0.execute_method.remote('warmup'),
                        worker1.execute_method.remote('warmup')]))

    num_layers = engine_config.model_config.get_num_layers(engine_config.parallel_config)
    head_size = engine_config.model_config.get_head_size()
    num_heads = engine_config.model_config.get_num_kv_heads(engine_config.parallel_config)
    block_size = engine_config.cache_config.block_size

    dummy_data = torch.randn(size=(num_layers, 2, num_gpu_blocks, block_size*num_heads*head_size))
    ray.get(worker0.execute_method.remote('set_gpu_cache', data=dummy_data))
    worker0_data = ray.get(worker0.execute_method.remote('get_gpu_cache'))

    dst_blocks = list(range(num_gpu_blocks))
    random.shuffle(dst_blocks)
    src_to_dst = dict(enumerate(dst_blocks))
    ray.get(worker1.execute_method.remote(
        'migrate_cache',
        src_worker_handle_list=[worker0],
        src_blocks=list(src_to_dst.keys()),
        dst_blocks=list(src_to_dst.values())))

    worker1_data = ray.get(worker1.execute_method.remote('get_gpu_cache'))

    for layer_idx in range(num_layers):
        for src_idx, dst_idx in src_to_dst.items():
            assert torch.allclose(worker0_data[layer_idx][0][src_idx], worker1_data[layer_idx][0][dst_idx])
            assert torch.allclose(worker0_data[layer_idx][1][src_idx], worker1_data[layer_idx][1][dst_idx])

    ray.shutdown()



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch
import ray
from ray.util.scheduling_strategies import NodeAffinitySchedulingStrategy

from vllm.engine.arg_utils import EngineArgs
from vllm.utils import get_distributed_init_method, get_ip, get_open_port
from vllm.worker.cache_engine import CacheEngine
from vllm.config import EngineConfig
from vllm.executor.ray_gpu_executor import RayWorkerWrapper

from llumnix.arg_utils import EngineManagerArgs
from llumnix.utils import random_uuid


def create_worker(rank: int, local_rank: int, engine_config: EngineConfig,
                  worker_module_name="llumnix.backends.vllm.worker",
                  worker_class_name="MigrationWorker"):
    scheduling_strategy = NodeAffinitySchedulingStrategy(
                node_id=ray.get_runtime_context().get_node_id(),
                soft=False,
            )

    worker = ray.remote(
        num_cpus=0,
        num_gpus=1,
        scheduling_strategy=scheduling_strategy
    )(RayWorkerWrapper).remote(
        worker_module_name=worker_module_name,
        worker_class_name=worker_class_name,
        trust_remote_code=True
    )

    worker.init_worker.remote(
        model_config=engine_config.model_config,
        parallel_config=engine_config.parallel_config,
        scheduler_config=engine_config.scheduler_config,
        device_config=engine_config.device_config,
        cache_config=engine_config.cache_config,
        load_config=engine_config.load_config,
        local_rank=local_rank,
        rank=rank,
        distributed_init_method=get_distributed_init_method(get_ip(), get_open_port()),
        lora_config=engine_config.lora_config,
        vision_language_config=engine_config.vision_language_config,
        is_driver_worker = False
    )

    return worker

@pytest.mark.parametrize("backend", ['rpc', 'gloo', 'nccl'])
def test_reserve_memory_for_migration(backend):
    ray.init(namespace="llumnix", ignore_reinit_error=True)

    engine_config = EngineArgs(model='facebook/opt-125m', max_model_len=8, enforce_eager=True).create_engine_config()
    migraiton_config = EngineManagerArgs(migration_cache_blocks=1).create_migration_config()
    migraiton_config.migration_backend = backend
    worker = create_worker(rank=0, local_rank=0, engine_config=engine_config)
    ray.get(worker.execute_method.remote('init_device'))

    block_size = CacheEngine.get_cache_block_size(engine_config.cache_config, engine_config.model_config,
                                                  engine_config.parallel_config)
    num_layers = engine_config.model_config.get_num_layers(engine_config.parallel_config)
    occupy_memory = migraiton_config.migration_cache_blocks * block_size * migraiton_config.migration_num_layers // num_layers

    migration_cache_size = ray.get(worker.execute_method.remote('reserve_memory_for_migration',
                                                                migration_config=migraiton_config,
                                                                model_config=engine_config.model_config,
                                                                cache_config=engine_config.cache_config,
                                                                parallel_config=engine_config.parallel_config))
    assert migration_cache_size == occupy_memory

    ray.shutdown()

@pytest.mark.skipif(torch.cuda.device_count() < 2, reason="Need at least 2 GPU to run the test.")
@pytest.mark.parametrize("backend", ['rpc', 'gloo', 'nccl'])
def test_rebuild_migration_backend(backend):
    ray.init(namespace="llumnix", ignore_reinit_error=True)

    engine_config = EngineArgs(model='facebook/opt-125m', max_model_len=8, enforce_eager=True).create_engine_config()
    migraiton_config = EngineManagerArgs(migration_cache_blocks=1).create_migration_config()
    migraiton_config.migration_backend = backend

    worker0 = create_worker(rank=0, local_rank=0, engine_config=engine_config)
    worker0_id = random_uuid()
    ray.get(worker0.execute_method.remote('init_device'))
    ray.get(worker0.execute_method.remote('initialize_cache', num_gpu_blocks=8, num_cpu_blocks=0))
    ray.get(worker0.execute_method.remote(
        'init_migration',
        instance_id=worker0_id,
        migration_config=migraiton_config,
        src_worker_handle_list=[worker0],
        node_id=ray.get_runtime_context().get_node_id()))
    instance_rank = {worker0_id: 0}
    assert ray.get(worker0.execute_method.remote('rebuild_migration_backend', instance_rank=instance_rank,
                                                 group_name=random_uuid()))
    assert ray.get(worker0.execute_method.remote('warmup'))

    worker1 = create_worker(rank=0, local_rank=0, engine_config=engine_config)
    worker1_id = random_uuid()
    ray.get(worker1.execute_method.remote('init_device'))
    ray.get(worker1.execute_method.remote('initialize_cache', num_gpu_blocks=8, num_cpu_blocks=0))
    ray.get(worker1.execute_method.remote(
        'init_migration',
        instance_id=worker1_id,
        migration_config=migraiton_config,
        src_worker_handle_list=[worker1],
        node_id=ray.get_runtime_context().get_node_id()))

    instance_rank = {worker1_id: 1, worker0_id: 0}
    group_name = random_uuid()
    assert all(ray.get([worker0.execute_method.remote('rebuild_migration_backend',
                                                      instance_rank=instance_rank, group_name=group_name),
                        worker1.execute_method.remote('rebuild_migration_backend',
                                                      instance_rank=instance_rank, group_name=group_name)]))
    assert all(ray.get([worker0.execute_method.remote('warmup'),
                        worker1.execute_method.remote('warmup')]))

    ray.kill(worker1)

    instance_rank = {worker0_id: 0}
    assert ray.get(worker0.execute_method.remote('rebuild_migration_backend', instance_rank=instance_rank,
                                                group_name=random_uuid()))
    assert ray.get(worker0.execute_method.remote('warmup'))

    ray.shutdown()



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
import torch
from func_timeout import func_set_timeout, FunctionTimedOut

import ray
import ray.util.collective as col
from vllm.worker.cache_engine import CacheEngine
from llumnix.config import MigrationConfig
from llumnix.backends.migration_backend_interface import MigrationBackendBase
from llumnix.logger import init_logger

logger = init_logger(__name__)

@ray.remote(num_cpus=0)
class ProxyActor:
    def exec_method(self, is_driver_worker, handle, *args, **kwargs):
        try:
            if is_driver_worker:
                ret = ray.get(handle.execute_engine_method.remote("execute_worker_method", *args, **kwargs))
            else:
                ret = ray.get(handle.execute_method.remote(*args, **kwargs))
        # pylint: disable=try-except-raise
        except:
            raise

        return ret

NUMPY_SUPPORTED_DTYPES = [torch.float32, torch.float16]

class RayRpcMigrationBackend(MigrationBackendBase):
    def __init__(self, migration_config: MigrationConfig, cache_engine: CacheEngine,  worker_rank, worker_handle_list, \
                  scheduling_strategy, is_driver_worker, gpu_cache) -> None:
        super().__init__()

        self.migration_config = migration_config
        self.cache_engine = cache_engine

        self.worker_rank = worker_rank
        self.worker_handle_list = worker_handle_list
        self.actor = ProxyActor.options(scheduling_strategy=scheduling_strategy).remote()

        self.rpc_dtype = self.cache_engine.dtype
        if self.cache_engine.dtype in NUMPY_SUPPORTED_DTYPES:
            self.rpc_dtype = self.cache_engine.dtype
        else:
            self.rpc_dtype = torch.float32
            logger.warning("Detect numpy unsupported dtype: {}. Using torch.float32.".format(self.cache_engine.dtype))

        self.is_driver_worker = is_driver_worker
        self.gpu_cache = gpu_cache
        self.cache_device = "cpu"
        self.num_migration_cache_blocks = self.migration_config.migration_cache_blocks
        self.num_layers = self.cache_engine.num_layers
        self.migration_cache_size = self.cache_engine.block_size * self.cache_engine.num_heads * self.cache_engine.head_size

        self.dummy_cache = torch.empty(
            size=(self.num_migration_cache_blocks, self.num_layers, 2, self.migration_cache_size),
            dtype=self.cache_engine.dtype,
            device=self.cache_device,
            pin_memory=True
        )
        self.migration_stream = torch.cuda.Stream()

    def init_backend(self, group_name, world_size, rank) -> bool:
        logger.info("create rpc migration backend successfully.")
        return True

    def destory_backend(self) -> None:
        # rpc migration backend does not need to be destroyed as there is no group.
        # It uses ray actor handle to migration cache blocks.
        pass

    def warmup(self) -> bool:
        self.actor.exec_method.remote(self.is_driver_worker, "do_send", [0])
        logger.info("rpc migration backend warmup successfully.")
        return True

    # The src actor will pack the kv-cache data layer by layer. Specifically, NumPy is used for the transfer
    # because, for a single node, Ray RPC can transfer NumPy arrays via shared memory. Then, the recv actor
    # first copies the data to a pinned-memory dummy cache before transferring it to the GPU to accelerate data transfer.
    def migrate_cache(self, src_handle, src_blocks: List[int], dst_blocks: List[int]) -> None:
        tot_blocks = len(src_blocks)
        rpc_numpy_cache = None
        for start_idx in range(0, tot_blocks, self.num_migration_cache_blocks):
            offset = min(self.num_migration_cache_blocks, tot_blocks - start_idx)
            send_blocks = src_blocks[start_idx:start_idx+offset]
            ray_obj = self.actor.exec_method.remote(self.is_driver_worker, src_handle, "do_send", None, send_blocks)
            if rpc_numpy_cache is not None:
                self.do_recv(rpc_numpy_cache, recv_blocks)
            rpc_numpy_cache = ray.get(ray_obj)
            recv_blocks = dst_blocks[start_idx:start_idx+offset]
        self.do_recv(rpc_numpy_cache, recv_blocks)

    def do_send(self, dst_handle, blocks: List[int]):
        num_blocks = len(blocks)
        send_cache = self.dummy_cache[:num_blocks].view(self.num_layers, 2, num_blocks, self.migration_cache_size)
        src_to_dst = {block_num: idx for idx, block_num in enumerate(blocks)}
        with torch.cuda.stream(self.migration_stream):
            for layer_idx in range(self.num_layers):
                self.cache_engine.attn_backend.swap_blocks(self.gpu_cache[layer_idx], send_cache[layer_idx], src_to_dst)
        torch.cuda.Stream.synchronize(self.migration_stream)
        return send_cache.to(self.rpc_dtype).numpy()

    def do_recv(self, src_handle, blocks: List[int]):
        num_blocks = len(blocks)
        src_to_dst = dict(enumerate(blocks))
        recv_cache = self.dummy_cache[:num_blocks].view(self.num_layers, 2, num_blocks, self.migration_cache_size)
        # use pin memory dummy_cache to speed up data transfer
        recv_cache.copy_(torch.from_numpy(src_handle))

        with torch.cuda.stream(self.migration_stream):
            for layer_idx in range(self.num_layers):
                self.cache_engine.attn_backend.swap_blocks(recv_cache[layer_idx], self.gpu_cache[layer_idx], src_to_dst)
        torch.cuda.Stream.synchronize(self.migration_stream)

def try_import_gloo():
    try:
        # pylint: disable=C0415
        from ray.util.collective.collective_group import gloo_util
        import pygloo

        # Add support for bf16 type in Gloo. Now bf16 and fp16 both map to glooFloat16, but this is okay because
        # Gloo only uses the data type size for transmission.
        # pylint: disable=W0221,I1101
        gloo_util.TORCH_GLOO_DTYPE_MAP[torch.bfloat16] = pygloo.glooDataType_t.glooFloat16
    except ImportError as e:
        raise ImportError("Gloo is not installed. Please install it first.") from e

class RayColMigrationBackend(MigrationBackendBase):
    def __init__(self, migration_config: MigrationConfig, cache_engine: CacheEngine, local_rank,
                 scheduling_strategy, is_driver_worker, gpu_cache) -> None:
        super().__init__()

        import cupy

        self.migration_config = migration_config
        self.cache_engine = cache_engine
        self.backend = migration_config.migration_backend
        self.migration_num_layers = min(migration_config.migration_num_layers, self.cache_engine.num_layers)
        self.num_migration_cache_blocks = migration_config.migration_cache_blocks

        self.backend = migration_config.migration_backend
        self.global_world_size = -1
        self.global_rank = -1
        self.group_name = None

        self.local_rank = local_rank
        self.actor = ProxyActor.options(scheduling_strategy=scheduling_strategy).remote()
        self.is_driver_worker = is_driver_worker
        self.gpu_cache = gpu_cache

        self.migration_cache_size = self.cache_engine.block_size * self.cache_engine.num_heads * self.cache_engine.head_size

        if self.backend == 'gloo':
            try_import_gloo()
            self.cache_device = "cpu"
        else:
            self.cache_device = torch.device(f"cuda:{self.local_rank}")

        pin_memory = (self.backend == 'gloo')
        self.dummy_cache = torch.empty(
            size=(self.num_migration_cache_blocks, self.migration_num_layers, 2, self.migration_cache_size),
            dtype=self.cache_engine.dtype,
            device=self.cache_device,
            pin_memory=pin_memory
        )

        self.migration_stream = cupy.cuda.Stream()

    def init_backend(self, group_name, world_size, rank) -> bool:
        @func_set_timeout(self.migration_config.migration_backend_init_timeout)
        def init_group(world_size, rank, backend, group_name):
            col.init_collective_group(world_size, rank, backend, group_name)

        try:
            init_group(world_size, rank, self.backend, group_name)
        except FunctionTimedOut:
            logger.info("create migration backend fail (group_name: {}, world_size: {}, rank: {}, backbend: {})."
                .format(group_name, world_size, rank, self.backend))
            return False

        self.group_name = group_name
        self.global_world_size = world_size
        self.global_rank = rank

        logger.info("create migration backend group successfully (group_name: {}, world_size: {}, rank: {}, backbend: {})."
                    .format(self.group_name, self.global_world_size, self.global_rank, self.backend))
        return True

    def destory_backend(self) -> None:
        if self.group_name is None:
            return

        err_info = None
        try:
            col.destroy_collective_group(self.group_name)
        # pylint: disable=W0703
        except Exception as e:
            err_info = e

        if err_info is not None:
            logger.info("destory migration backend successfully (group_name: {}, backbend: {}), error: {}."
                    .format(self.group_name, self.backend, err_info))
        else:
            logger.info("destory migration backend successfully (group_name: {}, backbend: {})."
                    .format(self.group_name, self.backend))

        self.group_name = None

    def warmup(self) -> bool:
        if self.global_world_size > 1:
            try:
                col.allreduce(self.dummy_cache[0], self.group_name)
            # pylint: disable=W0703
            except Exception as e:
                logger.info("warmup migration backend failed (group_name: {}, world_size: {}, rank: {}, backbend: {}), err: {}."
                    .format(self.group_name, self.global_world_size, self.global_rank, self.backend, e))
                return False

        logger.info("migration backend warmup successfully (group_name: {}, world_size: {}, rank: {}, backbend: {})."
                    .format(self.group_name, self.global_world_size, self.global_rank, self.backend))
        return True

    # Ray.collective is used to construct the gloo and nccl backends. The do_send/do_recv functions will transmit
    # data layer by layer. Take into consideration that col.send/recv are blocking operations.
    def migrate_cache(self, src_handle, src_blocks: List[int], dst_blocks: List[int]) -> None:
        tot_blocks = len(src_blocks)
        src_rank = ray.get(self.actor.exec_method.remote(self.is_driver_worker, src_handle, "get_global_rank"))

        for start_idx in range(0, tot_blocks, self.num_migration_cache_blocks):
            offset = min(self.num_migration_cache_blocks, tot_blocks - start_idx)
            send_blocks = src_blocks[start_idx:start_idx+offset]
            recv_blocks = dst_blocks[start_idx:start_idx+offset]
            self.actor.exec_method.remote(self.is_driver_worker, src_handle, "do_send", self.global_rank, send_blocks)
            self.do_recv(src_rank, recv_blocks)

    def do_send(self, dst_handle, blocks: List[int]):
        num_blocks = len(blocks)
        send_cache = self.dummy_cache[:num_blocks].view(self.migration_num_layers, 2, num_blocks, self.migration_cache_size)
        src_to_dst = {block_num: idx for idx, block_num in enumerate(blocks)}

        with self.migration_stream:
            for layer_idx in range(self.cache_engine.num_layers):
                cache_idx = layer_idx % self.migration_num_layers
                self.cache_engine.attn_backend.swap_blocks(self.gpu_cache[layer_idx], send_cache[cache_idx], src_to_dst)
                if cache_idx + 1 == self.migration_num_layers or layer_idx + 1 == self.cache_engine.num_layers:
                    # TODO(KuilongCui): check the error code if peer is dead
                    col.send(send_cache, dst_handle, self.group_name)
        self.migration_stream.synchronize()

    def do_recv(self, src_handle, blocks: List[int]):
        num_blocks = len(blocks)
        src_to_dst = dict(enumerate(blocks))
        recv_cache = self.dummy_cache[:num_blocks].view(self.migration_num_layers, 2, num_blocks, self.migration_cache_size)

        with self.migration_stream:
            for layer_idx in range(self.cache_engine.num_layers):
                cache_idx = layer_idx % self.migration_num_layers
                if cache_idx == 0:
                    col.recv(recv_cache, src_handle, self.group_name)
                self.cache_engine.attn_backend.swap_blocks(recv_cache[cache_idx], self.gpu_cache[layer_idx], src_to_dst)
        self.migration_stream.synchronize()

def get_migration_backend(migration_config: MigrationConfig, cache_engine: CacheEngine, worker_handle_list, scheduling_strategy,
                        is_driver_worker, gpu_cache, worker_rank, local_rank) -> MigrationBackendBase:
    if cache_engine.num_gpu_blocks < migration_config.migration_cache_blocks:
        logger.warning("migration_cache_blocks({}) is larger than num_gpu_blocks({}), reducing it to num_gpu_blocks."
                       .format(migration_config.migration_cache_blocks, cache_engine.num_gpu_blocks))
        migration_config.migration_cache_blocks = cache_engine.num_gpu_blocks

    target_col = None
    backend = migration_config.migration_backend
    if backend in ['nccl', 'gloo']:
        target_col = RayColMigrationBackend(migration_config, cache_engine, local_rank, scheduling_strategy,
                                            is_driver_worker, gpu_cache)
    else:
        target_col = RayRpcMigrationBackend(migration_config, cache_engine, worker_rank, worker_handle_list,
                                            scheduling_strategy, is_driver_worker, gpu_cache)

    return target_col



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import time
import csv
import os
from typing import Dict, List, Tuple, Union, Iterable
from collections import defaultdict
import traceback
import ray

from llumnix.llumlet.llumlet import Llumlet
from llumnix.logger import init_logger
from llumnix.global_scheduler.global_scheduler import GlobalScheduler
from llumnix.instance_info import InstanceInfo
from llumnix.config import GlobalSchedulerConfig
from llumnix.arg_utils import EngineManagerArgs
from llumnix.backends.profiling import ProfilingDatabase
from llumnix.server_info import ServerInfo
from llumnix.backends.backend_interface import BackendType
from llumnix.utils import random_uuid


logger = init_logger(__name__)

MANAGER_ACTOR_NAME = 'manager'
CLEARING_INTERVAL = 3600
RETRIES_INTERVALS = 5.0

# TODO(s5u13b): Fix the logger when manager failover.


class LLMEngineManager:
    def __init__(self,
                 engine_manager_args: EngineManagerArgs,
                 global_scheduler_config: GlobalSchedulerConfig,
                 work_dir: str,
                 log_requests: bool = True,
                 profiling_database: ProfilingDatabase = None) -> None:
        os.chdir(work_dir)
        self.actor_name = MANAGER_ACTOR_NAME
        self.engine_manager_args = engine_manager_args
        self.profiling_database = profiling_database

        self.log_requests = log_requests

        self.num_instances = 0
        self.enable_migration = engine_manager_args.enable_migration
        self.enable_scaling = engine_manager_args.enable_scaling
        self.max_instances = engine_manager_args.max_instances
        self.min_instances = engine_manager_args.min_instances

        logger.info("LLMEngineManager starts")
        logger.info("enable_migration: {}".format(self.enable_migration))
        logger.info("num_instances: {}".format(self.num_instances))
        logger.info("max_instances: {}, min_instances: {}".format(self.max_instances, self.min_instances))

        # TODO(s5u13b): refactor auto-scaling

        self.instances: Dict[str, Llumlet] = {}
        self.instance_migrating: Dict[str, bool] = {}
        self.pending_rebuild_migration_instances = 0
        self.global_scheduler = GlobalScheduler(global_scheduler_config)
        # When manager starts, it automatically connects to all existing instances.
        self._connect_to_instances()

        self.polling_interval = engine_manager_args.polling_interval
        asyncio.create_task(self._update_instance_info_loop(self.polling_interval))

        # args
        self.pair_migration_frequency = engine_manager_args.pair_migration_frequency
        self.scaling_interval = engine_manager_args.scaling_interval

        # request states
        self.request_instance: Dict[str, str] = {}
        self.clearing_interval = CLEARING_INTERVAL
        asyncio.create_task(self._clear_request_instance_loop(self.clearing_interval))

        # migrate states
        self.num_instance_info_updates = 0
        self.migrating = False

        # auto-scaling states
        self.scale_up_time = -1
        self.scale_down_time = -1
        self.scaling_up = False
        self.scaling_down = False
        self.last_check_scale_time = time.time() + 100

        self.log_instance_info = engine_manager_args.log_instance_info
        if self.log_instance_info:
            self._init_instance_info_csv(engine_manager_args)

    async def generate(
            self,
            request_id: str,
            server_info: ServerInfo,
            *args,
            **kwargs,) -> None:
        while self.num_instances == 0:
            logger.info("No instance available temporarily, sleep {}s, "
                        "and retry generate request {} again....".format(RETRIES_INTERVALS, request_id))
            await asyncio.sleep(RETRIES_INTERVALS)
        instance_id = self.global_scheduler.dispatch()
        try:
            await self.instances[instance_id].generate.remote(request_id, server_info, *args, **kwargs)
            if self.log_requests:
                logger.info("received request {}.".format(request_id))
                logger.info("dispath to instance {}".format(instance_id))
                self.request_instance[request_id] = instance_id
        except (ray.exceptions.RayActorError, KeyError):
            logger.info("[generate] instance {} is dead, regenerate request {}".format(instance_id, request_id))
            self.scale_down(instance_id)

    async def abort(self, request_id: Union[str, Iterable[str]]) -> None:
        if isinstance(request_id, str):
            request_id = (request_id,)
        request_ids = set(request_id)
        instance_requests = defaultdict(list)
        for req_id in request_ids:
            if req_id in self.request_instance:
                instance_id = self.request_instance[req_id]
                instance_requests[instance_id].append(req_id)
        for instance_id, request_ids in instance_requests.items():
            try:
                # Requests will be free by instance when finished, so it is acceptable to miss aborted requests.
                await self.instances[instance_id].abort.remote(request_ids)
                if self.log_requests:
                    logger.info("abort requests: {}.".format(request_ids))
                    for req_id in request_ids:
                        if req_id in self.request_instance:
                            del self.request_instance[req_id]
            except (ray.exceptions.RayActorError, KeyError):
                logger.info("[abort] instance {} is dead".format(instance_id))
                self.scale_down(instance_id)

    async def _get_request_instance(self) -> None:
        logger.info("_get_request_instance:")
        tasks = [instance_actor_handle.get_all_request_ids.remote() for instance_actor_handle in self.instances.values()]
        instance_ids = list(self.instances.keys())
        rets = await asyncio.gather(*tasks, return_exceptions=True)
        instance_requests = []
        instance_id_list = []
        for idx, ret in enumerate(rets):
            if not isinstance(ret, ray.exceptions.RayActorError):
                instance_requests.append(ret)
                instance_id_list.append(instance_ids[idx])
            else:
                instance_id = instance_ids[idx]
                logger.info("[_get_request_instance] instance {} is dead".format(instance_id))
                self.scale_down(instance_id)
        logger.info("instance_id_list: {}".format(instance_id_list))
        logger.info("instance_requests: {}".format(instance_requests))
        for (instance_id, requests) in zip(instance_id_list, instance_requests):
            for request_id in requests:
                self.request_instance[request_id] = instance_id

    async def _update_instance_info_loop(self, interval: float) -> None:
        while True:
            try:
                await asyncio.sleep(interval)
                tasks = [instance.get_instance_info.remote() for instance in self.instances.values()]
                instance_ids = list(self.instances.keys())
                rets = await asyncio.gather(*tasks, return_exceptions=True)

                instance_info_list = []
                dead_instances = []
                for idx, ret in enumerate(rets):
                    if not isinstance(ret, ray.exceptions.RayActorError):
                        if ret is not None:
                            instance_info_list.append(ret)
                    else:
                        instance_id = instance_ids[idx]
                        dead_instances.append(instance_id)
                if len(dead_instances) > 0:
                    logger.info("[_update_instance_info_loop] dead instances: {}.".format(dead_instances))
                    self.scale_down(dead_instances)

                self.global_scheduler.update_instance_infos(instance_info_list)
                self.num_instance_info_updates += 1
                # Push migrate when the instance_info have updated a certain number of times.
                if self.enable_migration and self.num_instance_info_updates != 0 \
                    and self.num_instance_info_updates % self.pair_migration_frequency == 0:
                    asyncio.create_task(self._migrate())
                if self.log_instance_info:
                    self._log_instance_infos_to_csv(instance_info_list)
            # pylint: disable=W0703
            except Exception as e:
                logger.error("unexpected exception occurs: {}".format(e))
                logger.error("exception traceback: {}".format(traceback.format_exc()))

    async def _clear_request_instance_loop(self, interval: float):
        await self._get_request_instance()
        # Clear the request_instance at a certain interval to prevent memory leaking.
        while True:
            await asyncio.sleep(interval)
            self.request_instance = {}

    async def _post_migrate(self, rets: List[str], call_migrate_instance_pairs: List[Tuple[str, str]]) -> None:
        for i, ret in enumerate(rets):
            self.instance_migrating[call_migrate_instance_pairs[i][0]] = False
            self.instance_migrating[call_migrate_instance_pairs[i][1]] = False
            if isinstance(ret, (ray.exceptions.RayActorError, KeyError)):
                has_error_pair = await self._check_instance_error(call_migrate_instance_pairs[i])
                for j, has_error in enumerate(has_error_pair):
                    # Instance without error should clear migration states.
                    if not has_error:
                        try:
                            await self.instances[call_migrate_instance_pairs[i][j]].clear_migration_states.remote(is_migrate_in=bool(j))
                        except (ray.exceptions.RayActorError, KeyError):
                            has_error = True
                for j, has_error in enumerate(has_error_pair):
                    if has_error:
                        instance_id = call_migrate_instance_pairs[i][j]
                        logger.info("[_migrate] instance {} is dead".format(instance_id))
                        self.scale_down(instance_id)
            else:
                migrate_out_request_ids = ret
                if migrate_out_request_ids:
                    migrate_out_request_id = migrate_out_request_ids[0]
                    self.request_instance[migrate_out_request_id] = call_migrate_instance_pairs[i][1]
                logger.info("{}->{} migrate done, migrate request {}".format(
                    call_migrate_instance_pairs[i][0], call_migrate_instance_pairs[i][1], migrate_out_request_ids))

    async def _migrate(self) -> None:
        migrate_instance_pairs = self.global_scheduler.pair_migration()
        try:
            migration_tasks = []
            call_migrate_instance_pairs: List[Tuple[str, str]] = []
            for _, migrate_instance_pair in enumerate(migrate_instance_pairs):
                migrate_out_instance_id, migrate_in_instance_id = migrate_instance_pair
                if self.instance_migrating[migrate_out_instance_id] or self.instance_migrating[migrate_in_instance_id]:
                    continue
                self.instance_migrating[migrate_out_instance_id] = True
                self.instance_migrating[migrate_in_instance_id] = True
                migrate_in_instance_name = "instance_{}".format(migrate_in_instance_id)
                logger.info("{}->{} begin migrate out".format(migrate_out_instance_id, migrate_in_instance_id))
                call_migrate_instance_pairs.append(migrate_instance_pair)
                task = self.instances[migrate_out_instance_id].migrate_out.remote(migrate_in_instance_name)
                migration_tasks.append(task)
            # TODO(s5u13b): Migration failover could be implemented in Llumlet rather than manager.
            rets = await asyncio.gather(*migration_tasks, return_exceptions=True)
            await self._post_migrate(rets, call_migrate_instance_pairs)
        # pylint: disable=W0703
        except Exception as e:
            logger.error("unexpected exception occurs: {}".format(e))
            logger.error("exception traceback: {}".format(traceback.format_exc()))

    async def rebuild_migrate_backend(self) -> None:
        # Wait for all instances to finish migration
        while any(self.instance_migrating.values()):
            await asyncio.sleep(0.1)

        # During rebuilding migration backend, disable migrate
        origin_config = self.enable_migration
        self.enable_migration = False

        async def run_task(alive_instances: List[str], task_name: str, *args, **kwargs):
            tasks = []
            for instance_name in alive_instances:
                llumlet_handle = self.instances[instance_name]
                tasks.append(llumlet_handle.execute_engine_method.remote("_run_workers", task_name, *args, **kwargs))

            rets = await asyncio.gather(*tasks, return_exceptions=True)

            dead_instances = set()
            for instance_name, ret in zip(alive_instances, rets):
                if isinstance(ret, ray.exceptions.RayActorError):
                    dead_instances.add(instance_name)

            if len(dead_instances) > 0:
                self.scale_down(dead_instances, rebuild_migrate_backend=False)

                if self.engine_manager_args.migration_backend == 'gloo':
                    try:
                        # clear gloo migrate backend intermediate state
                        ray.kill(ray.get_actor("gloo_queue", "llumnix"))
                    except ValueError:
                        # gloo_queue may not have been created yet; just ignore this error.
                        pass

            return dead_instances

        alive_instances = sorted(self.instances.keys())
        pending_task = self.pending_rebuild_migration_instances
        group_name = None

        while len(alive_instances) > 0 and self.pending_rebuild_migration_instances > 0:
            dead_instances = set()
            group_name = random_uuid()
            instance_rank = {instance_id: index for index, instance_id in enumerate(alive_instances)}
            dead_instances.update(await run_task(alive_instances, "rebuild_migration_backend",
                                                                  instance_rank, group_name))

            if len(dead_instances) == 0 and self.pending_rebuild_migration_instances == pending_task:
                dead_instances.update(await run_task(alive_instances, "warmup"))

            if len(dead_instances) == 0:
                self.pending_rebuild_migration_instances -= pending_task

            alive_instances = sorted(set(self.instances.keys()) - dead_instances)
            pending_task = self.pending_rebuild_migration_instances

        if len(alive_instances) == 0:
            self.pending_rebuild_migration_instances = 0
            group_name = None

        logger.info("rebuild {} migrate backend done, group_name: {}, alive instance ({}): {}"
            .format(self.engine_manager_args.migration_backend, group_name, len(alive_instances), alive_instances))

        # Restore migrate config
        self.enable_migration = origin_config

    def scale_up(self, instance_id: Union[str, Iterable[str]], llumlet_actor_handles: List["ray.actor.ActorHandle"]) -> None:
        if isinstance(instance_id, str):
            instance_id = [instance_id,]
        instance_ids = list(instance_id)

        indeed_update = False
        no_pending_instance = (self.pending_rebuild_migration_instances == 0)

        for idx, ins_id in enumerate(instance_ids):
            if ins_id not in self.instances:
                indeed_update = True
                self.instances[ins_id] = llumlet_actor_handles[idx]
                self.instance_migrating[ins_id] = False
                self.pending_rebuild_migration_instances += 1
        self.global_scheduler.scale_up(instance_ids)
        self.num_instances = len(self.instances)

        # When scaling up, we need to rebuild the migration backend. But if initially self.pending_rebuild_migration_instances != 0,
        # a coroutine is already handling the changes in the number of instances in the cluster and it will account for the changes
        # caused by this scale-up (see rebuild_migrate_backend for details). Therefore, we simply return in this case. Specifically,
        # for RPC, the Ray actor handle is used for the migration cache, so there is no need to rebuild the group.
        if self.engine_manager_args.migration_backend != 'rpc' and indeed_update and no_pending_instance:
            asyncio.create_task(self.rebuild_migrate_backend())

        return self.num_instances

    def scale_down(self, instance_id: Union[str, Iterable[str]], rebuild_migrate_backend: bool = True) -> None:
        if isinstance(instance_id, str):
            instance_id = [instance_id,]
        instance_ids = list(instance_id)

        indeed_update = False
        no_pending_instance = self.pending_rebuild_migration_instances == 0

        for ins_id in instance_ids:
            if ins_id in self.instances:
                indeed_update = True
                del self.instances[ins_id]
                del self.instance_migrating[ins_id]
                self.pending_rebuild_migration_instances += 1
        self.global_scheduler.scale_down(instance_ids)
        self.num_instances = len(self.instances)

        if self.engine_manager_args.migration_backend != 'rpc':
            if len(self.instances) == 0:
                self.pending_rebuild_migration_instances = 0
            elif indeed_update and no_pending_instance and rebuild_migrate_backend:
                asyncio.create_task(self.rebuild_migrate_backend())

        return self.num_instances

    def _connect_to_instances(self):
        actor_names_dict = ray.util.list_named_actors(True)
        instance_actor_names = [actor_name_dict['name'] for actor_name_dict in actor_names_dict if actor_name_dict['name'] != MANAGER_ACTOR_NAME]
        instance_actor_handles = [ray.get_actor(actor_name, namespace='llumnix') for actor_name in instance_actor_names]
        scale_up_instance_ids = []
        scale_up_instance_actor_handles = []
        for instance_actor_name, instance_actor_handle in zip(instance_actor_names, instance_actor_handles):
            instance_id = instance_actor_name[len('instance_'):]
            if instance_id not in self.instances:
                try:
                    ray.get(instance_actor_handle.is_ready.remote())
                # pylint: disable=W0703
                except Exception as e:
                    logger.info("connect to instance {} abort, which may be not ready or alive, err: {}".format(instance_id, e))
                    continue
                logger.info("connect to instance {}.".format(instance_id))
                scale_up_instance_ids.append(instance_id)
                scale_up_instance_actor_handles.append(instance_actor_handle)
        # The only function that can add instance actor handles to manager.
        self.scale_up(scale_up_instance_ids, scale_up_instance_actor_handles)

    async def _check_instance_error(self, migrate_instance_pairs: Tuple[str, str]) -> List[bool]:
        results = [None, None]
        for idx, instance_id in enumerate(migrate_instance_pairs):
            try:
                await self.instances[instance_id].is_ready.remote()
                logger.info("[_check_instance_error] instance {} is alive".format(instance_id))
                results[idx] = False
            except (ray.exceptions.RayActorError, KeyError):
                logger.info("[_check_instance_error] instance {} is dead".format(instance_id))
                results[idx] = True
        return results

    @classmethod
    def from_args(cls,
                  engine_manager_args: EngineManagerArgs,
                  profiling_database: ProfilingDatabase=None) -> "LLMEngineManager":
        global_scheduler_config = engine_manager_args.create_global_scheduler_configs()
        # Init manager actor in 'llumnix' namespace to ensure that only one manager can be created.
        manager_class = ray.remote(num_cpus=0,
                                   max_restarts=-1,
                                   name=MANAGER_ACTOR_NAME,
                                   namespace='llumnix',
                                   lifetime="detached"
                                   )(cls)
        engine_manager = manager_class.remote(engine_manager_args,
                                              global_scheduler_config,
                                              os.getcwd(),
                                              log_requests=not engine_manager_args.disable_log_requests_manager,
                                              profiling_database=profiling_database)
        logger.info("engine_manager_args: {}".format(engine_manager_args))
        return engine_manager

    # TODO(s5u13b): Significant duplication with llumlet_utils.init_llumlets. Consider reducing duplicate codes.
    # TODO(s5u13b): Fix the logger when enabling init instance by manager.
    def init_llumlets(self,
                      engine_args,
                      node_id: str) -> Tuple[List[str], List[Llumlet]]:
        engine_manager_args = self.engine_manager_args
        engine_config = engine_args.create_engine_config()
        parallel_config = engine_config.parallel_config
        instance_ids: List[str] = []
        llumlets: List[Llumlet] = []
        for _ in range(engine_manager_args.initial_instances):
            instance_id = random_uuid()
            if not engine_manager_args.profiling_result_file_path:
                llumlet = Llumlet.from_args(
                    engine_manager_args.disable_fixed_node_init_instance,
                    True,
                    node_id,
                    instance_id,
                    BackendType.VLLM,
                    parallel_config.world_size,
                    engine_manager_args.create_migration_config(),
                    engine_args,
                )
            else:
                llumlet = Llumlet.from_args(
                    engine_manager_args.disable_fixed_node_init_instance,
                    True,
                    node_id,
                    instance_id,
                    BackendType.SIM_VLLM,
                    parallel_config.world_size,
                    engine_manager_args.create_migration_config(),
                    engine_manager_args.profiling_result_file_path,
                    engine_manager_args.gpu_type,
                    engine_args,
                )
            instance_ids.append(instance_id)
            llumlets.append(llumlet)

        return instance_ids, llumlets

    def get_actor_name(self) -> str:
        return self.actor_name

    async def is_ready(self) -> bool:
        """Called by api server, return true when all the instances have been successfully created."""
        tasks = [llumlet.is_ready.remote() for llumlet in self.instances.values()]
        is_ready_list = await asyncio.gather(*tasks)
        return all(is_ready_list)

    def _init_instance_info_csv(self, engine_manager_args: EngineManagerArgs) -> None:
        # pylint: disable=consider-using-with
        self.instance_info_file = open(engine_manager_args.log_filename + '_instance.csv', 'w', encoding='utf-8')
        self.instance_info_csv = csv.writer(self.instance_info_file)
        self.instance_info_csv.writerow([
            'timestamp',
            'instance_id',
            'step_id',
            'gpu_cache_usage',
            'num_available_gpu_blocks',
            'instance_load',
            'max_tot_tokens',
            'num_running_requests',
            'num_waiting_requests',
            'num_killed_requests',
            'inference_type',
            'bs',
            'latency',
            'seq_lens',
            'num_instances',
            'num_seqs',
            'num_blocks_first_waiting_request',
            'num_blocks_all_waiting_requests',
            'waiting_time_first_waiting_request'])

    def _log_instance_infos_to_csv(self, instance_infos: List[InstanceInfo]) -> None:
        for instance_info in instance_infos:
            self.instance_info_csv.writerow([
                instance_info.timestamp,
                instance_info.instance_id,
                instance_info.step_id,
                instance_info.gpu_cache_usage,
                instance_info.num_available_gpu_blocks,
                instance_info.instance_load_migrate,
                instance_info.max_tot_tokens,
                instance_info.num_running_requests,
                instance_info.num_waiting_requests,
                instance_info.num_killed_requests,
                instance_info.inference_type,
                instance_info.num_batched_tokens,
                instance_info.latency,
                instance_info.running_seq_lens,
                self.num_instances,
                instance_info.num_seqs,
                instance_info.num_blocks_first_waiting_request,
                instance_info.num_blocks_all_waiting_requests,
                instance_info.waiting_time_first_waiting_request])
        self.instance_info_file.flush()



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Iterable, List, Optional, Union

from llumnix.llumlet.migrating_request import MigratingRequest
from llumnix.server_info import ServerInfo


class BackendType(str, Enum):
    VLLM = "VLLM"
    SIM_VLLM = "SIM_VLLM"

    @staticmethod
    def is_sim_backend(status: "BackendType") -> bool:
        return status in [
            BackendType.SIM_VLLM,
        ]

class BackendInferenceType(str, Enum):
    PREFILL = "prefill"
    DECODE = "decode"

class BackendInterface(ABC):
    # Methods for inference
    @abstractmethod
    def add_request(self, request_id: str, server_info: ServerInfo,
                    *args, **kwargs) -> None:
        """Adds a new inference request to the backend's processing queue.

        This method should capture all necessary metadata of an inference request
        such as request ID, arrival time, and any other pertinent information.

        Args:
            request_id: Request ID.
            server_info: The information of the api server where the request come.
            *args: Positional arguments that represent request-specific data.
            **kwargs: Keyword arguments that contain metadata of the backend request
                      (request_id, arrival_time, etc.).
        """
        raise NotImplementedError

    @abstractmethod
    def abort_request(self, request_id: Union[str, Iterable[str]]) -> None:
        """Aborts one or more requests currently being processed.

        Args:
            request_id: A single request ID or an iterable of request IDs to abort.
        """
        raise NotImplementedError

    @abstractmethod
    def _start_engine_loop(self) -> None:
        """Start step loop of backend engine.
        """
        raise NotImplementedError

    # Methods for migration
    @abstractmethod
    def get_request_incremental_blocks(self, backend_request: Any, pre_stage_num_blocks: int) -> List[int]:
        """Retrieves the incremental block table for a given request.

        This method is used to fetch a list of block numbers that represent the incremental
        data associated with a particular backend request. It is typically called during a
        migration process where only transfer incremental blocks since last migration stage.

        Args:
            backend_request: An object representing the backend request. The type of this
                             object is dependent on the backend implementation and the details
                             of the request.
            pre_stage_num_blocks: An integer representing the number of blocks that were
                                   transferred in the previous migration stages.
                                   This is used to determine the incremental blocks that
                                   need to be fetched in the current stage.

        Returns:
            A list of integers, where each integer represents a block number that indicates
            physical index of kv cache block tensor. These block numbers can then be used
            to transfer to dstination instance.
        """
        raise NotImplementedError

    @abstractmethod
    def remove_running_request(self, request_id: str) -> None:
        """
        Removes a request from the backend's running queue.

        This method is responsible for safely halting and removing an active request from the running
        queue of the backend engine. This action is performed in last stage migration when
        a request should be suspend.

        Args:
            request_id: A string identifier for the request that is to be removed from the running
                        queue. This ID uniquely identifies the request within the backend system.
        """
        raise NotImplementedError

    @abstractmethod
    def add_migrating_out_request_last_stage(self, backend_request: Any) -> None:
        """
        Adds a backend request to the list of migrating out request in last stage.

        This method adds a backend request to the list of migrating out request in last stage. This action is performed after
        the migrating out request has been removed from the running queue.

        Args:
            backend_request: An object representing the backend request. The type of this
                             object is dependent on the backend implementation and the details
                             of the request.
        """
        raise NotImplementedError

    @abstractmethod
    def remove_migrating_out_request_last_stage(self, backend_request: Any) -> None:
        """
        Removes a backend request from the list of migrating out request in last stage.

        This method adds a backend request to the list of migrating out request in last stage. This action is performed after
        the migration is finished successfully.

        Args:
            backend_request: An object representing the backend request. The type of this
                             object is dependent on the backend implementation and the details
                             of the request.
        """
        raise NotImplementedError

    @abstractmethod
    def pop_migrating_out_requests_last_stage(self) -> List[Any]:
        """
        Pops the list of migrating out request in last stage.

        This method pops the list of migrating out request in last stage. This action is performed to free migrating out requests
        in last stage when the migration encounters exception.

        Returns:
            The list of migrating out request in last stage.
        """
        raise NotImplementedError

    @abstractmethod
    def pre_alloc(self, request_id: str, block_num: int) -> List[int]:
        """Pre-allocates cache blocks for a migrating request.

        This method selects a specified number of free cache blocks to be reserved for an incoming
        migration request identified by the given request ID. It updates the pre-allocation cache
        dictionary with the allocated blocks, which ensures that these blocks are not used by
        another process until the migration is finished.

        Args:
            request_id: The unique identifier of the migration request for which cache blocks
                        are to be pre-allocated.
            block_num: The number of cache blocks that need to be pre-allocated for the request.

        Returns:
            A list of integers where each integer represents the block table reserved for the migration request.
        """
        raise NotImplementedError

    @abstractmethod
    def should_abort_migration(self, backend_request: Any, last_stage_time: int) -> bool:
        """
        Determines whether the migration for a specific backend request should be aborted.

        This method evaluates the conditions under which a migration process, associated with a
        backend request, should be terminated before completion. If the backend request is no longer valid or a preemption
        event has occurred since the last migration stage, the migration need to be aborted.

        Args:
            backend_request: An object representing the backend request. The type of this
                             object is dependent on the backend implementation and the details
                             of the request.
            last_stage_time: An integer timestamp representing the last successful stage of the
                             migration process. This is used to determine if any significant event
                             has occurred after this point that would warrant the abortion of the
                             migration.

        Returns:
            True if the migration should be aborted based on the evaluation of the backend_request
            and last_stage_time; False otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def add_running_request(self, backend_request: Any) -> None:
        """
        Adds a backend request to the running queue for processing.

        This method enqueues a backend request into engine running queue, marking it for
        active processing. It is used when a suspend migrating request should be added back
        to running queue.

        Args:
            backend_request: An object representing the backend request. The type of this
                             object is dependent on the backend implementation and the details
                             of the request.
        """
        raise NotImplementedError

    @abstractmethod
    def is_request_running(self, backend_request: Any) -> bool:
        """Checks if a given backend request is currently in the running queue.

        This method determines whether a backend request is present and actively being processed
        in the running queue.

        Args:
            backend_request: An object representing the backend request. The type of this
                             object is dependent on the backend implementation and the details
                             of the request.

        Returns:
            True if the backend request is currently in the running queue; False otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def free_dst_pre_alloc_cache(self, request_id: str = None) -> None:
        """Frees pre-allocated blocks for a migrating request on the destination instance.

        This method is responsible for releasing any cache blocks or other resources that were
        pre-allocated on the destination instance for a migrating request. This is typically called
        after the migration aborted to ensure that resources are not left reserved unnecessarily.

        Args:
            request_id: A string representing the unique identifier of the request for which resources
                        are to be freed on the destination instance.
        """
        raise NotImplementedError

    @abstractmethod
    def free_src_request(self, backend_request: Any) -> None:
        """Frees blocks associated with a migrating request on the source instance.

        Upon completion or cancellation of a migration process, this method is invoked to clean up and
        release any reserved resources such as block tables and other metadata on the source instance.

        Args:
            backend_request: An object representing the backend request. The type of this
                             object is dependent on the backend implementation and the details
                             of the request.
        """
        raise NotImplementedError

    @abstractmethod
    def send_blocks(self, dst_ray_actor: "ray.actor.ActorHandle", src_blocks: List[int], dst_blocks: List[int]):
        """
        Sends cache blocks from the source instance to the destination instance.

        This method orchestrates the physical transfer of cache blocks between instances by Ray. It is responsible
        for ensuring that the specified blocks from the source instance's cache are sent to and properly received by
        the destination instance, where they are mapped according to the destination block table.

        Args:
            dst_ray_actor: A handle to the Ray actor representing the destination instance where the cache
                           blocks are to be sent. This handle is used to reference the destination's
                           execution context and manage the block transfer.
            src_blocks: A list of integers representing the block indexs in the source instance's
                             cache that need to be sent to the destination.
            dst_blocks: A list of integers representing the block indexs in the destination instance's
                             cache where the incoming blocks should be stored.
        """
        raise NotImplementedError

    @abstractmethod
    def commit_dst_request(self, backend_request: Any, server_info: ServerInfo) -> None:
        """Commits the migrating request to the destination instance.

        This method finalizes the migration process by transferring all necessary metadata and resource
        information (such as the block table) to the destination instance's engine. Upon completion,
        the destination instance should have all the required data to resume and handle the request as if it originated there natively.

        Args:
            backend_request: An object representing the backend request. The type of this
                             object is dependent on the backend implementation and the details
                             of the request.
            server_info: The information of the api server where the request come.
        """
        raise NotImplementedError

    @abstractmethod
    def get_last_running_request(self) -> Optional[MigratingRequest]:
        """Retrieves the last non-prefilling request from the running queue.

        This method iterates over the running queue in reverse order and returns the last request
        that has moved past the prefilling stage. Prefilling requests are not considered for
        migration, as they have not yet begun processing and therefore do not have any state that
        needs to be preserved.

        Returns:
            An instance of MigratingRequest representing the last request in the running queue that
            is not prefilling, or None if there are no such requests in the queue.
        """
        raise NotImplementedError

    @abstractmethod
    def get_longest_running_request(self) -> Optional[MigratingRequest]:
        """Retrieves the request with the longest sequence length from the running queue.

        This method should sorts the running queue based on length of the requests and
        returns the non-prefilling longest one.

        Returns:
            An instance of MigratingRequest representing the longest-running request in the queue that
            has generated output, or None if no such request exists.
        """
        raise NotImplementedError

    @abstractmethod
    def get_shortest_running_request(self) -> Optional[MigratingRequest]:
        """Retrieves the request with the shortest sequence length from the running queue.

        This method should sorts the running queue based on length of the requests and
        returns the non-prefilling shortest one.

        Returns:
            An instance of MigratingRequest representing the shortest-running request in the queue that
            has generated output, or None if no such request exists.
        """
        raise NotImplementedError

    @abstractmethod
    def get_request_server_info(self, request_id: str) -> ServerInfo:
        """Retrieves the information of the api server where the request come.

        This method is used by the migration coordinator to get the information of server where the migrating request come.

        Args:
            request_id: Request ID.
        Returns:
            The request output queue of the api server where the request come.
        """
        raise NotImplementedError

    @abstractmethod
    def get_all_request_ids(self) -> List[str]:
        """Retrieves all requests in instance.

        This method is used by the manager to get all requests in instance when it restarts.

        Returns:
            The list of request ID.
        """
        raise NotImplementedError

    @abstractmethod
    def free_request_states(self, request_id: Union[str, Iterable[str]]) -> None:
        """Free request states recorded in backend engine.

        This method is used by the llumlet or backend engine to free the request states when the request is finished/migrated/aborted.

        Args:
            request_id: Single/List of request ID.
        """
        raise NotImplementedError



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import vllm
from vllm import *

from llumnix.server_info import ServerInfo
from llumnix.entrypoints.llumnix_utils import (launch_ray_cluster, connect_to_ray_cluster,
                                               init_manager, init_llumlets)
from llumnix.arg_utils import EngineManagerArgs
from llumnix.llm_engine_manager import LLMEngineManager
from llumnix.llumlet.llumlet import Llumlet

from .version import __version__

__all__ = [
    "__version__",
    "ServerInfo",
    "launch_ray_cluster",
    "connect_to_ray_cluster",
    "init_manager",
    "init_llumlets",
    "EngineManagerArgs",
    "LLMEngineManager",
    "Llumlet"
]

__all__.extend(getattr(vllm, "__all__", []))



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class MigrationConfig:
    def __init__(
            self,
            request_migration_policy: str,
            migration_backend: str,
            migration_cache_blocks: int,
            migration_num_layers: int,
            last_stage_max_blocks: int,
            max_stages: int,
            migration_backend_init_timeout: float) -> None:
        self.request_migration_policy = request_migration_policy
        self.migration_backend = migration_backend
        self.migration_num_layers = migration_num_layers
        self.migration_cache_blocks = migration_cache_blocks
        self.last_stage_max_blocks = last_stage_max_blocks
        self.max_stages = max_stages
        self.migration_backend_init_timeout = migration_backend_init_timeout

class GlobalSchedulerConfig:
    def __init__(
            self,
            initial_instances: int,
            load_metric: str,
            dispatch_policy: str,
            pair_migration_policy: str,
            migrate_out_threshold: float,
            enable_defrag: bool,
            scaling_policy: str,
            scale_up_threshold: float,
            scale_down_threshold: float) -> None:
        self.initial_instances = initial_instances
        self.load_metric = load_metric

        self.dispatch_policy = dispatch_policy

        self.pair_migration_policy = pair_migration_policy
        self.migrate_out_load_threshold = migrate_out_threshold*(-1)
        self.enable_defrag = enable_defrag

        self.scaling_policy = scaling_policy
        self.scale_up_threshold = scale_up_threshold*(-1)
        self.scale_down_threshold = scale_down_threshold*(-1)



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import sys
import os
import time
from typing import List, Tuple
import asyncio
import ray

from ray.util.queue import Queue as RayQueue
from ray.util.scheduling_strategies import NodeAffinitySchedulingStrategy

from llumnix.llm_engine_manager import LLMEngineManager, MANAGER_ACTOR_NAME
from llumnix.llumlet.llumlet import Llumlet
from llumnix.backends.backend_interface import BackendType
from llumnix.logger import init_logger
from llumnix.utils import random_uuid
from llumnix.arg_utils import EngineManagerArgs


logger = init_logger(__name__)

# TODO(s5u13b): Set the values through tests.
MAX_RESTARTS = 30
RESTART_INTERVALS = 1
MAX_TASK_RETRIES = 300
RETRIES_INTERVALS = 0.1


def get_ip_address():
    result = subprocess.run(['hostname', '-i'], stdout=subprocess.PIPE, check=True)
    ip_address = result.stdout.decode('utf-8').strip()
    return ip_address

def launch_ray_cluster(ray_cluster_port: int) -> subprocess.CompletedProcess:
    head_node_ip = os.getenv('HEAD_NODE_IP')
    node_ip_address = get_ip_address()
    try:
        # Stop the existing ray processes on the node first.
        subprocess.run(['ray', 'stop', '--force'], check=True, text=True, capture_output=True)
    except subprocess.CalledProcessError as e:
        logger.info("'ray stop' failed with: \n{}".format(e.stderr))
        sys.exit(1)
    # Need to specify the head node ip through environment variable currently.
    if head_node_ip is None:
        logger.info("Environment variable 'HEAD_NODE_IP' should be set for ray cluster launch.")
        sys.exit(1)
    ray_start_command = None
    if 'HEAD_NODE' in os.environ:
        ray_start_command = f"ray start --head --node-ip-address={node_ip_address} --port={ray_cluster_port}"
        try:
            result = subprocess.run(['ray', 'start', '--head', f'--port={ray_cluster_port}'], check=True, text=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            logger.info("'{}' failed with: \n{}".format(ray_start_command, e.stderr))
            sys.exit(1)
    else:
        ray_start_command = f"ray start --address={head_node_ip}:{ray_cluster_port} --node-ip-address={node_ip_address}"
        for attempt in range(MAX_RESTARTS):
            try:
                # wait about 2 mins by default
                result = subprocess.run(['ray', 'start', f'--address={head_node_ip}:{ray_cluster_port}'], check=True, text=True, capture_output=True)
                break
            except subprocess.CalledProcessError as e:
                if attempt < MAX_RESTARTS:
                    print("Execute '{}' repeatedly until the head node starts...".format(ray_start_command))
                    time.sleep(RESTART_INTERVALS)
                else:
                    logger.info("'{}' failed after {} attempts with: \n{}".format(ray_start_command, attempt, e.stderr))
                    sys.exit(1)
    logger.info("'{}' succeeed with: \n{}".format(ray_start_command, result.stdout))
    return result

def connect_to_ray_cluster(port: int, namespace="llumnix") -> None:
    head_node_ip = os.getenv('HEAD_NODE_IP')
    ray.init(address=f"{head_node_ip}:{port}", ignore_reinit_error=True, namespace=namespace)

def is_gpu_available() -> bool:
    try:
        subprocess.check_output(['nvidia-smi'])
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False

def retry_manager_method_sync(ray_call, method_name, *args, **kwargs):
    for attempt in range(MAX_TASK_RETRIES):
        try:
            ret = ray.get(ray_call(*args, **kwargs))
            break
        except ray.exceptions.RayActorError:
            if attempt < MAX_TASK_RETRIES - 1:
                logger.info("Manager is unavailable, sleep {}s, and retry {} again...".format(RETRIES_INTERVALS, method_name))
                time.sleep(RETRIES_INTERVALS)
            else:
                logger.info("After {} times retries, manager is still unavailable".format(MAX_TASK_RETRIES))
                raise
    return ret

async def retry_manager_method_async(ray_call, method_name, *args, **kwargs):
    for attempt in range(MAX_TASK_RETRIES):
        try:
            ret = await ray_call(*args, **kwargs)
            break
        except ray.exceptions.RayActorError:
            if attempt < MAX_TASK_RETRIES - 1:
                logger.info("Manager is unavailable, sleep {}s, and retry {} again...".format(RETRIES_INTERVALS, method_name))
                await asyncio.sleep(RETRIES_INTERVALS)
            else:
                logger.info("After {} times retries, manager is still unavailable".format(MAX_TASK_RETRIES))
                raise
    return ret

def init_manager(engine_manager_args: EngineManagerArgs) -> LLMEngineManager:
    # Only one instance create the manager actor, the other instances get the existing manager actor through ray.
    try:
        engine_manager = LLMEngineManager.from_args(engine_manager_args, None)
        logger.info("Init LLMEngineManager on current node")
    except ValueError:
        engine_manager = ray.get_actor(MANAGER_ACTOR_NAME, namespace='llumnix')
        logger.info("Get existing LLMEngineManager")
    return engine_manager

def init_llumlets(engine_manager_args: EngineManagerArgs,
                  engine_args,
                  node_id: str) -> Tuple[List[str], List[Llumlet]]:
    engine_config = engine_args.create_engine_config()
    parallel_config = engine_config.parallel_config
    instance_ids: List[str] = []
    llumlets: List[Llumlet] = []

    instance_ids = [random_uuid() for _ in range(engine_manager_args.initial_instances)]
    migration_configs = engine_manager_args.create_migration_config()

    for idx in range(engine_manager_args.initial_instances):
        instance_id = instance_ids[idx]
        if not engine_manager_args.profiling_result_file_path:
            llumlet = Llumlet.from_args(
                engine_manager_args.disable_fixed_node_init_instance,
                False,
                node_id,
                instance_id,
                BackendType.VLLM,
                parallel_config.world_size,
                migration_configs,
                engine_args,
            )
        else:
            llumlet = Llumlet.from_args(
                engine_manager_args.disable_fixed_node_init_instance,
                False,
                node_id,
                instance_id,
                BackendType.SIM_VLLM,
                parallel_config.world_size,
                migration_configs,
                engine_manager_args.profiling_result_file_path,
                engine_manager_args.gpu_type,
                engine_args,
            )
        llumlets.append(llumlet)
    return instance_ids, llumlets

def init_request_output_queue() -> RayQueue:
    # request_output_queue should be placed in the same node as the api server.
    request_output_queue = RayQueue(actor_options={
        "scheduling_strategy": NodeAffinitySchedulingStrategy(
            node_id=ray.get_runtime_context().get_node_id(),
            soft=False,)
    })
    return request_output_queue

def init_llumnix_components(engine_manager_args: EngineManagerArgs,
                            engine_args,
                            node_id: str) -> Tuple[LLMEngineManager, List[Llumlet], RayQueue]:
    engine_manager = init_manager(engine_manager_args)
    if engine_manager_args.disable_init_instance_by_manager:
        instance_ids, llumlets = init_llumlets(engine_manager_args, engine_args, node_id)
    else:
        instance_ids, llumlets = retry_manager_method_sync(engine_manager.init_llumlets.remote, 'init_llumlets', engine_args, node_id)

    available_instance_ids = []
    dead_instance_ids = []
    available_llumlets = []

    ready_tasks = [llumlet.is_ready.remote() for llumlet in llumlets]

    for idx, task in enumerate(ready_tasks):
        try:
            ray.get(task)
            available_instance_ids.append(instance_ids[idx])
            available_llumlets.append(llumlets[idx])
        except ray.exceptions.RayActorError:
            dead_instance_ids.append(instance_ids[idx])

    if len(dead_instance_ids) > 0:
        retry_manager_method_sync(engine_manager.scale_down.remote, 'scale_down', dead_instance_ids)

    if len(available_instance_ids) > 0:
        retry_manager_method_sync(engine_manager.scale_up.remote, 'scale_up',
                                  available_instance_ids, available_llumlets)
        logger.info("Init Llumnix components done, {} instances are ready, instance_ids: {}."
                    .format(len(available_instance_ids), available_instance_ids))

    request_output_queue = init_request_output_queue()

    return engine_manager, available_instance_ids, available_llumlets, request_output_queue



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
import ray

from llumnix.arg_utils import EngineManagerArgs
from llumnix.entrypoints.llumnix_utils import (get_ip_address,
                                               launch_ray_cluster,
                                               init_manager,
                                               init_request_output_queue,
                                               retry_manager_method_sync,
                                               retry_manager_method_async)
from llumnix.llm_engine_manager import MANAGER_ACTOR_NAME


def test_launch_ray_cluster():
    ip_address = get_ip_address()
    os.environ['HEAD_NODE'] = '1'
    os.environ['HEAD_NODE_IP'] = ip_address
    result = launch_ray_cluster(30050)
    assert result.returncode == 0

def test_init_manager():
    engine_manager_args = EngineManagerArgs()
    engine_manager = init_manager(engine_manager_args)
    assert engine_manager is not None
    engine_manager_actor_handle = ray.get_actor(MANAGER_ACTOR_NAME, namespace='llumnix')
    assert engine_manager_actor_handle is not None
    assert engine_manager == engine_manager_actor_handle
    ray.kill(engine_manager)
    ray.shutdown()

def test_init_request_output_queue():
    request_output_queue = init_request_output_queue()
    assert request_output_queue is not None
    ray.shutdown()

def test_retry_manager_method_sync():
    engine_manager_args = EngineManagerArgs()
    engine_manager = init_manager(engine_manager_args)
    ret = retry_manager_method_sync(engine_manager.is_ready.remote, 'is_ready')
    assert ret is True
    ray.kill(engine_manager)
    ray.shutdown()

@pytest.mark.asyncio
async def test_retry_manager_method_async():
    engine_manager_args = EngineManagerArgs()
    engine_manager = init_manager(engine_manager_args)
    ret = await retry_manager_method_async(engine_manager.is_ready.remote, 'is_ready')
    assert ret is True
    ray.kill(engine_manager)
    ray.shutdown()



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List


class MigrationBackendBase(ABC):
    @abstractmethod
    def init_backend(self, group_name, world_size, rank) -> bool:
        raise NotImplementedError

    @abstractmethod
    def destory_backend(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def warmup(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def migrate_cache(self, src_handle, src_blocks: List[int], dst_blocks: List[int]) -> None:
        raise NotImplementedError

    @abstractmethod
    def do_send(self, dst_handle, blocks: List[int]):
        raise NotImplementedError

    @abstractmethod
    def do_recv(self, src_handle, blocks: List[int]):
        raise NotImplementedError



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ray.util.queue import Queue as RayQueue


class ServerInfo:
    def __init__(self,
                 server_id: str,
                 request_output_queue: RayQueue) -> None:
        self.server_id = server_id
        self.request_output_queue = request_output_queue



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asyncio.log import logger
import time
import threading
from typing import Dict, List, Optional, Tuple

from vllm.sequence import SequenceGroup
from vllm.core.block_manager_v1 import BlockSpaceManagerV1, BlockTable
from vllm.core.scheduler import (Scheduler, SequenceGroupMetadata, SchedulerOutputs,
                                 PreemptionMode, SequenceStatus)

from llumnix.instance_info import InstanceInfo
from llumnix.logger import init_logger
from llumnix.llumlet.local_migration_scheduler import MigratingRequest
from llumnix.backends.backend_interface import BackendInferenceType
from llumnix.backends.vllm.utils import scheduler_lock

logger = init_logger(__name__)


# TODO(ZeldaHuang): adapt prefix cache and sliding window, now use v1 manager
class BlockManagerLlumnix(BlockSpaceManagerV1):
    def get_free_blocks(self, num_required_blocks: int) -> BlockTable:
        num_free_gpu_blocks = self.gpu_allocator.get_num_free_blocks()
        if (num_free_gpu_blocks - num_required_blocks <
                self.watermark_blocks):
            return []
        blocks = []
        for _ in range(num_required_blocks):
            block = self.gpu_allocator.allocate()
            # Set the reference counts of the token blocks.
            block.ref_count = 1
            blocks.append(block)
        return blocks

    def add_block_table(self, block_table: BlockTable, seq_id: int) -> None:
        self.block_tables[seq_id] = block_table.copy()

class SchedulerLlumnix(Scheduler):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.block_manager = BlockManagerLlumnix(
            block_size=self.cache_config.block_size,
            num_gpu_blocks=self.cache_config.num_gpu_blocks,
            num_cpu_blocks=self.cache_config.num_cpu_blocks,
            sliding_window=self.cache_config.sliding_window,
            enable_caching=self.cache_config.enable_prefix_caching)
        self.pre_alloc_cache_dict: Dict[str, BlockTable] = {}
        self.last_preemption_time_dict: Dict[str, float] = {}
        self.prefilling_seq_groups = []
        self.scheduler_lock = threading.Lock()
        self.migrating_out_request_last_stage = []

    def add_update_instance_info_callback(self, update_instance_info_callback):
        self.update_instance_info_callback = update_instance_info_callback
        self.update_instance_info_callback(self._get_instance_info())

    def _preempt(
        self,
        seq_group: SequenceGroup,
        blocks_to_swap_out: Dict[int, int],
        preemption_mode: Optional[PreemptionMode] = None,
    ) -> PreemptionMode:
        self.last_preemption_time_dict[seq_group.request_id] = time.time()
        return super()._preempt(seq_group, blocks_to_swap_out, preemption_mode)

    def _get_num_killed_requests(self) -> int:
        cnt = len(self.swapped)
        for seq_group in self.waiting:
            if seq_group.request_id in self.last_preemption_time_dict:
                cnt += 1
        return cnt

    @scheduler_lock
    def get_last_running_request(self) -> Optional[MigratingRequest]:
        for seq_group in reversed(self.running):
            if seq_group not in self.prefilling_seq_groups:
                return MigratingRequest(seq_group.request_id, seq_group)
        return None

    @scheduler_lock
    def get_longest_running_request(self) -> Optional[MigratingRequest]:
        # TODO(ZeldaHuang): use for loop find request
        sorted_running = sorted(self.running, key=lambda seq_group: seq_group.get_seqs()[0].get_len())
        for seq_group in reversed(sorted_running):
            if seq_group not in self.prefilling_seq_groups:
                return MigratingRequest(seq_group.request_id, seq_group)
        return None

    @scheduler_lock
    def get_shortest_running_request(self) -> Optional[MigratingRequest]:
        sorted_running = sorted(self.running, key=lambda seq_group: seq_group.get_seqs()[0].get_len())
        for seq_group in sorted_running:
            if seq_group not in self.prefilling_seq_groups:
                return MigratingRequest(seq_group.request_id, seq_group)
        return None

    @scheduler_lock
    def get_all_request_ids(self) -> List[str]:
        request_ids : List[str] = []
        for state_queue in [self.waiting, self.running, self.swapped]:
            for seq_group in state_queue:
                request_ids.append(seq_group.request_id)
        return request_ids

    @scheduler_lock
    def get_request_incremental_blocks(self, backend_request: SequenceGroup, pre_stage_num_blocks: int) -> List[int]:
        seq = backend_request.get_seqs()[0]
        blocks = self.block_manager.get_block_table(seq)
        return blocks[pre_stage_num_blocks:]

    @scheduler_lock
    def remove_running_request(self, request_id: str) -> None:
        for seq_group in self.running:
            if seq_group.request_id == request_id:
                seq = seq_group.get_seqs()[0]
                self.running.remove(seq_group)
                seq.status = SequenceStatus.WAITING
                break

    def add_migrating_out_request_last_stage(self, backend_request: SequenceGroup) -> None:
        self.migrating_out_request_last_stage.append(backend_request)

    def remove_migrating_out_request_last_stage(self, backend_request: SequenceGroup) -> None:
        self.migrating_out_request_last_stage.remove(backend_request)

    def pop_migrating_out_requests_last_stage(self) -> List[SequenceGroup]:
        migrating_out_request_last_stage = self.migrating_out_request_last_stage.copy()
        self.migrating_out_request_last_stage.clear()
        return migrating_out_request_last_stage

    @scheduler_lock
    def pre_alloc(self, request_id: str, block_num: int) -> List[int]:
        blocks = self.block_manager.get_free_blocks(block_num)
        pre_blocks = self.pre_alloc_cache_dict.get(request_id, [])
        pre_blocks.extend(blocks)
        self.pre_alloc_cache_dict[request_id] = pre_blocks
        blocks = [block.block_number for block in blocks]
        return blocks

    @scheduler_lock
    def should_abort_migration(self, backend_request: SequenceGroup, last_stage_time: int) -> bool:
        if backend_request not in self.running or not backend_request.get_seqs()[0].get_output_len():
            # not in running queue or in prefill stage
            return True
        last_preemption_time = self.last_preemption_time_dict.get(backend_request.request_id, 0)
        if last_preemption_time > last_stage_time:
            return True
        return False

    @scheduler_lock
    def add_running_request(self, backend_request: SequenceGroup) -> None:
        seq = backend_request.get_seqs()[0]
        seq.status = SequenceStatus.RUNNING
        self.running.append(backend_request)

    @scheduler_lock
    def is_request_running(self, backend_request: SequenceGroup) -> bool:
        return backend_request in self.running

    @scheduler_lock
    def free_dst_pre_alloc_cache(self, request_id: str = None) -> None:
        if request_id:
            blocks = self.pre_alloc_cache_dict.pop(request_id, [])
            # pylint: disable=protected-access
            self.block_manager._free_block_table(blocks)
        else:
            # Clear all pre-allocated cache of dst instance when src instance encounters exception.
            request_ids = list(self.pre_alloc_cache_dict.keys())
            for req_id in request_ids:
                blocks = self.pre_alloc_cache_dict.pop(req_id, [])
                # pylint: disable=protected-access
                self.block_manager._free_block_table(blocks)

    @scheduler_lock
    def free_src_request(self, backend_request: SequenceGroup) -> None:
        seq = backend_request.get_seqs()[0]
        logger.info("free seq {}".format(seq.seq_id))
        self.free_seq(seq)

    def _get_instance_info(self) -> InstanceInfo:
        num_total_gpu_blocks = self.cache_config.num_gpu_blocks
        num_free_gpu_blocks = self.block_manager.get_num_free_gpu_blocks()
        num_used_gpu_blocks = num_total_gpu_blocks - num_free_gpu_blocks
        gpu_cache_usage = num_used_gpu_blocks / num_total_gpu_blocks
        if self.waiting:
            num_blocks_waiting_requests = []
            waiting_time_waiting_requests = []
            for seq_group in self.waiting:
                num_prompt_tokens = seq_group.get_seqs()[0].get_len()
                num_blocks = num_prompt_tokens / self.cache_config.block_size
                waiting_time = time.time() - seq_group.metrics.arrival_time
                num_blocks_waiting_requests.append(num_blocks)
                waiting_time_waiting_requests.append(waiting_time)
            num_blocks_first_waiting_request = num_blocks_waiting_requests[0]
            waiting_time_first_waiting_request = waiting_time_waiting_requests[0]
            num_blocks_all_waiting_requests = sum(num_blocks_waiting_requests)
        else:
            num_blocks_first_waiting_request = 0
            waiting_time_first_waiting_request = 0
            num_blocks_all_waiting_requests = 0
        instance_info = InstanceInfo(
            num_total_gpu_blocks=num_total_gpu_blocks,
            num_watermark_blocks=self.block_manager.watermark_blocks,
            num_used_gpu_blocks=num_used_gpu_blocks,
            num_free_gpu_blocks=num_free_gpu_blocks,
            gpu_cache_usage=gpu_cache_usage,
            num_running_requests=len(self.running),
            num_waiting_requests=len(self.waiting),
            num_killed_requests=self._get_num_killed_requests(),
            num_blocks_first_waiting_request=num_blocks_first_waiting_request,
            waiting_time_first_waiting_request=waiting_time_first_waiting_request,
            num_blocks_all_waiting_requests=num_blocks_all_waiting_requests,
            inference_type=BackendInferenceType.PREFILL if self.prefilling_seq_groups \
                           else BackendInferenceType.DECODE,
        )
        for seq_group in self.running:
            instance_info.running_seq_lens.extend([seq.get_len() for seq in seq_group.get_seqs()])
            instance_info.num_seqs = len(instance_info.running_seq_lens)
        # TODO(s5u13b): Only correct when using prefill preemption batching policy.
        instance_info.num_batched_tokens = sum([seq_group.get_seqs()[0].get_len() for seq_group in self.prefilling_seq_groups])\
                                         if self.prefilling_seq_groups else len(instance_info.running_seq_lens)
        instance_info.finished_request_ids = [seq_group.request_id for seq_group in self.running if seq_group.is_finished()]
        return instance_info

    @scheduler_lock
    def schedule(self) -> Tuple[List[SequenceGroupMetadata], SchedulerOutputs]:
        # record prefilling requests
        seq_group_metadata_list, scheduler_outputs = super().schedule()
        self.prefilling_seq_groups = []
        for scheduled_seq_group in scheduler_outputs.scheduled_seq_groups:
            if scheduled_seq_group.seq_group.is_prefill():
                self.prefilling_seq_groups.append(scheduled_seq_group.seq_group)
        self.update_instance_info_callback(self._get_instance_info())
        return seq_group_metadata_list, scheduler_outputs

    @scheduler_lock
    def add_seq_group(self, *args, **kwargs):
        return super().add_seq_group(*args, **kwargs)

    @scheduler_lock
    def abort_seq_group(self, *args, **kwargs):
        return super().abort_seq_group(*args, **kwargs)



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Profile the running time and memory usage of models"""
import os
import dataclasses

from collections import namedtuple
from typing import List, Dict, Any
from scipy.optimize import curve_fit

import cloudpickle
import pandas as pd
import numpy as np

from llumnix.backends.backend_interface import BackendType, BackendInferenceType

# 2D parallel configuration
# (gpu, tensor parallel, pipeline parallel)
SimParallelConfig = namedtuple("SimParallelConfig", ("gpu", "tp", "pp"))
# vllm blocks gpu cache configuration
SimCacheConfig = namedtuple("SimCacheConfig", ("gpu_memory_utilization", "block_size", "max_num_batched_tokens"))

def _pad_to_alignment(x, multiple_of):
    return x + ((-1*x) % multiple_of)

@dataclasses.dataclass
class LatencyMemData:
    # The latency of each stage
    # Type: Dict[tot_seq_len -> List[stage_latency]]
    prefill_latency: Dict
    # Type: Dict[(batch_size, tot_seq_len) -> List[stage_latency]]
    decode_latency: Dict
    # Type: Dict[SimCacheConfig -> num_gpu_blocks]
    cache_dict: Dict
    # Type: bandwidth(GB/s)
    migration_bandwidth: float = 1.0
    # Type: bandwidth(GB/s)
    swap_bandwidth: float = 32.0
    # Metadata for parallel strategy
    metadata: Any = None
    # model params for fitting latency
    decode_model_params: Any = None
    prefill_model_params: Any = None

    def add_latency_result(self, inference_type: BackendInferenceType, batch_size: int, tot_seq_len: int, latency: List[float]):
        if inference_type == BackendInferenceType.PREFILL:
            self.prefill_latency[batch_size] = latency
        else:
            self.decode_latency[(batch_size, tot_seq_len)] = latency

    def add_cache_result(self, cache_config: SimCacheConfig, num_blocks):
        self.cache_dict[cache_config] = num_blocks

    def get_prefill_dict_kv(self):
        return map(list, zip(*self.prefill_latency.items()))

    def get_decode_dict_kv(self):
        return map(list, zip(*self.decode_latency.items()))

@dataclasses.dataclass
class ProfilingResult:
    """Store the profiling result of a model."""
    model_name: str
    # The latency and memory usage of each pipeline stage on GPUs.
    para_dict: Dict[SimParallelConfig, LatencyMemData]
    # The latency of preprocess on CPU.
    preprocess_cpu: float = 0.0
    # The latency of postprocess on CPU.
    postprocess_cpu: float = 0.0

    def add_latency_result(self, parallel_config: SimParallelConfig, inference_type: BackendInferenceType, batch_size: int,
                           tot_seq_len: int, stage_latency: List[float], metadata: Any = None):
        """Add or overwrite the profiling results of a model."""
        if parallel_config not in self.para_dict:
            self.para_dict[parallel_config] = LatencyMemData(
                metadata=metadata, prefill_latency={}, decode_latency={}, cache_dict={})
            if inference_type == BackendInferenceType.PREFILL:
                self.para_dict[parallel_config].prefill_latency = {tot_seq_len: stage_latency}
            else:
                self.para_dict[parallel_config].decode_latency = {(batch_size, tot_seq_len): stage_latency}
        else:
            self.para_dict[parallel_config].add_latency_result(inference_type, batch_size, tot_seq_len, stage_latency)

    def add_cache_result(self, parallel_config: SimParallelConfig, cache_config: SimCacheConfig, num_blocks: int,
                         metadata: Any = None):
        if parallel_config not in self.para_dict:
            self.para_dict[parallel_config] = LatencyMemData(
                metadata=metadata, prefill_latency={}, decode_latency={}, cache_dict={})
        self.para_dict[parallel_config].add_cache_result(cache_config, num_blocks)

    def fit_from_database(self, parallel_config: SimParallelConfig):
        # fit prefill
        tot_seq_len_list, stage_latency_list = self.para_dict[parallel_config].get_prefill_dict_kv()
        latency_list = [stage_latency[0] for stage_latency in stage_latency_list]
        x = np.array(tot_seq_len_list)
        y = np.array(latency_list)
        params, _ = curve_fit(model_prefill, x, y)
        self.para_dict[parallel_config].prefill_model_params = params
        avg_loss = 0
        for idx, seq_len in enumerate(tot_seq_len_list):
            sim_lat = model_prefill(seq_len, *params)
            avg_loss += abs(sim_lat - latency_list[idx])
        print(f"prefill sim avg_loss={avg_loss/len(latency_list)}")
        # fit decode
        keys, stage_latency_list = self.para_dict[parallel_config].get_decode_dict_kv()
        bs_list, tot_seq_len_list = map(list, zip(*keys))
        latency_list = [stage_latency[0] for stage_latency in stage_latency_list]
        x = np.vstack((np.array(bs_list), np.array(tot_seq_len_list)))
        y = np.array(latency_list)
        params, _ = curve_fit(model_decode, x, y)
        avg_loss = 0
        self.para_dict[parallel_config].decode_model_params = params
        for idx, (bs, seq_len) in enumerate(zip(bs_list, tot_seq_len_list)):
            sim_lat = model_decode((bs, seq_len), *params)
            avg_loss += abs(sim_lat - latency_list[idx])
        print(f"decode sim avg_loss={avg_loss/len(latency_list)}")

class ProfilingDatabase:
    """Store the profiling results of all the models"""
    def __init__(self, database_filename: str, new_database: bool = False):
        # The file that backs up the profiling results.
        self.database_filename = database_filename
        # Dict[model_name -> ProfilingResult]
        self.results: Dict[str, ProfilingResult] = {}
        if not new_database:
            with open(database_filename, "rb") as f:
                self.results = cloudpickle.load(f)

    def get(self, model_name: str) -> ProfilingResult:
        return self.results.get(model_name)

    def update(self, result: ProfilingResult):
        self.results[result.model_name] = result

    def _extract_data(self, row):
        """Extract the profiling results from a row of the profiling CSV file."""
        inference_type = BackendInferenceType.PREFILL if row["inference_type"] == "prefill" else BackendInferenceType.DECODE
        # assert pp==1
        stage_latencies = [float(row["latency"])]
        batch_size = _pad_to_alignment(int(row["bs"]), 8)
        tot_seq_len = 0
        seq_lens_str = row["seq_lens"].strip('"[]"').split(",")
        for len_str in seq_lens_str:
            if len_str != "":
                tot_seq_len += int(len_str)
        tot_seq_len = _pad_to_alignment(tot_seq_len, 8)
        return stage_latencies, inference_type, batch_size, tot_seq_len

    def update_from_instance_log(self, file_name: str, model: str, parallel_config: SimParallelConfig):
        df = pd.read_csv(file_name+"_instance.csv")
        df = df[df['bs'] > 0]
        # read lines
        if model not in self.results:
            self.results[model] = ProfilingResult(model, {})
        for _, row in df.iterrows():
            stage_latencies, inference_type, batch_size, tot_seq_len = self._extract_data(row)
            self.results[model].add_latency_result(parallel_config, inference_type, batch_size, tot_seq_len, stage_latencies)

    def materialize(self):
        """Write the profiling results to the database file."""
        with open(self.database_filename, "wb") as f:
            cloudpickle.dump(self.results, f)

def model_prefill(x, a, b):
    return a * x + b

def model_decode(x, a, b, c):
    bs, tot_seq_len = x
    return a * bs + b * tot_seq_len + c

def get_latency_mem(backend_type: BackendType, profiling_database: ProfilingDatabase, gpu_type: str, **backend_args):
    assert BackendType.is_sim_backend(backend_type)
    if backend_type == BackendType.SIM_VLLM:
        # TODO(ZeldaHuang): support multi-lora, more device, vision language model
        model_config = backend_args.get("model_config")
        _ = backend_args.get("cache_config")
        parallel_config = backend_args.get("parallel_config")
        _ = backend_args.get("scheduler_config")
        model_name = model_config.model
        # get model_name from model path
        if model_name.endswith('/'):
            model_name = model_name[:-1]
        model_name = os.path.basename(model_name)
        profiling_result: ProfilingResult = profiling_database.get(model_name)
        sim_parallel_config = SimParallelConfig(gpu_type, parallel_config.tensor_parallel_size,
                                                parallel_config.pipeline_parallel_size)
        assert sim_parallel_config in profiling_result.para_dict.keys(), "sim parallel config not in database"
        latency_mem: LatencyMemData = profiling_result.para_dict[sim_parallel_config]
        return latency_mem
    raise ValueError(f'Unsupported backend: {backend_type}')

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--database", type=str, default="profiling.pkl")
    parser.add_argument("--log-path", type=str)
    parser.add_argument("--model", type=str)
    parser.add_argument("--gpu", type=str, default="a10")
    parser.add_argument("--tp", type=int, default=1)
    parser.add_argument("--pp", type=int, default=1)
    parser.add_argument("--gpu-memory-utilization", type=float, default=0.9)
    parser.add_argument("--block-size", type=int, default=16)
    parser.add_argument("--max-num-batched-tokens", type=int, default=8000)
    parser.add_argument("--num-gpu-blocks", type=int, default=0)
    parser.add_argument("--new-data", action="store_true")
    parser.add_argument("--fit", action="store_true")

    args = parser.parse_args()
    args_parallel_config = SimParallelConfig(args.gpu, args.tp, args.pp)
    args_cache_config = SimCacheConfig(args.gpu_memory_utilization, args.block_size, args.max_num_batched_tokens)
    database = ProfilingDatabase(args.database, args.new_data)
    if args.log_path:
        database.update_from_instance_log(args.log_path, args.model, args_parallel_config)
    if args.fit:
        model_result = database.get(args.model)
        model_result.fit_from_database(parallel_config=args_parallel_config)
    if args.num_gpu_blocks:
        model_result = database.get(args.model)
        model_result.add_cache_result(args_parallel_config, args_cache_config, args.num_gpu_blocks)

    database.materialize()



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Tuple, Set
from abc import ABC, abstractmethod
import numpy as np

from llumnix.logger import init_logger
from llumnix.instance_info import InstanceInfo, InstanceLoadCalculator

logger = init_logger(__name__)


class ScalingScheduler:
    def __init__(self,
                 scale_up_threshold: float,
                 scale_down_threshold: float,
                 scaling_policy: str,
                 instance_load_calculator: InstanceLoadCalculator) -> None:
        self.scale_up_threshold = scale_up_threshold
        self.scale_down_threshold = scale_down_threshold
        self.scaling_policy = ScalePolicyFactory.get_policy(scaling_policy,
                                                          instance_load_calculator=instance_load_calculator)
        self.instance_load_calculator = instance_load_calculator

        self.num_instances = 0
        self.instance_id_set: Set[str] = set()
        # instance info args
        self.instance_info: Dict[str, InstanceInfo] = None
        self.sorted_instance_infos: List[InstanceInfo] = None

    def check_scale(self) -> Tuple[str, str]:
        scale_up_num = 0
        scale_down_num = 0
        # if not all instances have returned instance_info, not scale
        if len(self.instance_info.keys()) < self.num_instances:
            return scale_up_num, scale_down_num
        now_instances = [self.instance_info[instance_id] for instance_id in self.instance_id_set]
        load_metric_up = self.scaling_policy.compute_load_metric_up(now_instances)
        load_metric_down = self.scaling_policy.compute_load_metric_down(now_instances)
        if load_metric_up > self.scale_up_threshold:
            while self.scaling_policy.compute_load_metric_avg(now_instances) > self.scale_up_threshold:
                scale_up_num += 1
                now_instances.append(self.get_empty_instance_info())
        elif load_metric_down < self.scale_down_threshold:
            scale_down_num = 1
        return scale_up_num, scale_down_num

    def update_instance_infos(self,
                              instance_info: Dict[str, InstanceInfo]) -> None:
        self.instance_info = instance_info

    def add_instance(self, instance_id: str) -> None:
        self.instance_id_set.add(instance_id)
        self.num_instances = len(self.instance_id_set)

    def remove_instance(self, instance_id: str) -> None:
        self.instance_id_set.remove(instance_id)
        self.num_instances = len(self.instance_id_set)

    def get_empty_instance_info(self) -> InstanceInfo:
        dummy_intance_info = InstanceInfo()
        dummy_intance_info.instance_id = -1
        dummy_intance_info.step_id = -1
        # TODO(s5u13b): Should be changed for proactive auto-scaling.
        dummy_intance_info.num_total_gpu_blocks = np.inf
        dummy_intance_info.num_available_gpu_blocks = np.inf
        dummy_intance_info.num_free_gpu_blocks = np.inf
        dummy_intance_info.num_available_gpu_blocks_waiting = np.inf
        return dummy_intance_info

class ScalePolicy(ABC):
    def __init__(self,
                 instance_load_calculator: InstanceLoadCalculator) -> None:
        self.instance_load_calculator = instance_load_calculator

    @abstractmethod
    def compute_load_metric_up(self, instance_infos: List[InstanceInfo]) -> float:
        pass

    @abstractmethod
    def compute_load_metric_down(self, instance_infos: List[InstanceInfo]) -> float:
        pass

    def compute_load_metric_avg(self, instance_infos: List[InstanceInfo]) -> float:
        tot_instance_info = InstanceInfo()
        tot_instance_info.instance_id = -1
        tot_instance_info.step_id = -1
        tot_instance_info.num_running_requests = sum([i.num_running_requests for i in instance_infos])
        tot_instance_info.num_waiting_requests = sum([i.num_waiting_requests for i in instance_infos])
        tot_instance_info.num_free_gpu_blocks = sum([i.num_free_gpu_blocks for i in instance_infos])
        tot_instance_info.num_total_gpu_blocks = sum([i.num_total_gpu_blocks for i in instance_infos])
        tot_instance_info.num_watermark_blocks = sum([i.num_watermark_blocks for i in instance_infos])
        tot_instance_info.num_blocks_all_waiting_requests = sum([i.num_blocks_all_waiting_requests for i in instance_infos])
        tot_instance_info.num_available_gpu_blocks = tot_instance_info.num_free_gpu_blocks - tot_instance_info.num_watermark_blocks
        return self.instance_load_calculator.compute_instance_load(tot_instance_info, action="scale")

class MaxLoad(ScalePolicy):
    def compute_load_metric_up(self, instance_infos: List[InstanceInfo]) -> float:
        return max([i.instance_load_dispatch_scale for i in instance_infos])

    def compute_load_metric_down(self, instance_infos: List[InstanceInfo]) -> float:
        return max([i.instance_load_dispatch_scale for i in instance_infos])
class MinLoad(ScalePolicy):
    def compute_load_metric_up(self, instance_infos: List[InstanceInfo]) -> float:
        return min([i.instance_load_dispatch_scale for i in instance_infos])

    def compute_load_metric_down(self, instance_infos: List[InstanceInfo]) -> float:
        return min([i.instance_load_dispatch_scale for i in instance_infos])

class AvgLoad(ScalePolicy):
    def compute_load_metric_up(self, instance_infos: List[InstanceInfo]) -> float:
        return self.compute_load_metric_avg(instance_infos)

    def compute_load_metric_down(self, instance_infos: List[InstanceInfo]) -> float:
        num_instances = len(instance_infos)
        tot_instance_info = InstanceInfo()
        tot_instance_info.instance_id = -1
        tot_instance_info.step_id = -1
        # the average load after scale down the last instance
        tot_instance_info.num_running_requests = sum([i.num_running_requests for i in instance_infos])
        tot_instance_info.num_waiting_requests = sum([i.num_waiting_requests for i in instance_infos])
        tot_instance_info.num_free_gpu_blocks = sum([i.num_free_gpu_blocks - i.num_total_gpu_blocks
                                                    if i.instance_id + 1 == num_instances else i.num_free_gpu_blocks
                                                    for i in instance_infos])
        tot_instance_info.num_free_gpu_blocks = max(0, tot_instance_info.num_free_gpu_blocks)
        tot_instance_info.num_total_gpu_blocks = sum([0 if i.instance_id + 1 == num_instances else i.num_total_gpu_blocks
                                                    for i in instance_infos])
        tot_instance_info.num_watermark_blocks = sum([0 if i.instance_id + 1 == num_instances else i.num_watermark_blocks
                                                    for i in instance_infos])
        tot_instance_info.num_blocks_all_waiting_requests = sum([i.num_blocks_all_waiting_requests for i in instance_infos])
        tot_instance_info.num_available_gpu_blocks = tot_instance_info.num_free_gpu_blocks - tot_instance_info.num_watermark_blocks
        return self.instance_load_calculator.compute_instance_load(tot_instance_info, action='scale')

class ScalePolicyFactory:
    _POLICY_REGISTRY = {
        'max_load': MaxLoad,
        'min_load': MinLoad,
        'avg_load': AvgLoad,
    }

    @classmethod
    def get_policy(cls, policy_name: str, **kwargs) -> ScalePolicy:
        return cls._POLICY_REGISTRY[policy_name](**kwargs)



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from llumnix.llumlet.migrating_request import MigratingRequest
from llumnix.backends.backend_interface import BackendInterface

class LocalMigrationScheduler:
    def __init__(self, request_migration_policy: str, backend_engine: BackendInterface) -> None:
        self.request_migration_policy = request_migration_policy
        self.backend_engine = backend_engine

    def get_migrate_out_request(self) -> Optional[MigratingRequest]:
        # TODO(s5u13b): remove the if-else codes
        migrate_out_request: MigratingRequest = None
        if self.request_migration_policy == 'LCFS':
            migrate_out_request = self.backend_engine.get_last_running_request()
        elif self.request_migration_policy in ['SJF', 'LJF']:
            if self.request_migration_policy == 'LJF':
                migrate_out_request = self.backend_engine.get_longest_running_request()
            elif self.request_migration_policy == 'SJF':
                migrate_out_request = self.backend_engine.get_shortest_running_request()
        return migrate_out_request
